@@ -1,3 +1,4 @@
+from black import Path
 import yaml
 import os
 from hubit.model import HubitModel
@@ -5,9 +6,11 @@ from hubit.model import HubitModel
 THISPATH = os.path.dirname(os.path.realpath(__file__))
 
 
-def get_model(filename, input_file="input.yml", model_id=""):
+def get_model(filename, input_file, model_id=""):
     hmodel = HubitModel.from_file(
-        os.path.join(THISPATH, filename), name=f"tank_{model_id}", output_path="./tmp"
+        os.path.join(THISPATH, filename),
+        name=f"tank_{model_id}_{Path(input_file).stem}",
+        output_path="./tmp",
     )
 
     # Load the input
