@@ -37,6 +37,7 @@ class PathIndexRange(str):
     supported ranges comprise
 
     - Positive integers e.g. `0`, `17`.
+    - Negative integers e.g. `-1`, `-2`.
     - The all-index character `:`.
 
     Further, in the `index_scope` attribute of
@@ -63,12 +64,23 @@ class PathIndexRange(str):
         else:
             return False
 
-    def __new__(self, value, allow_limited_range: bool = True):
+    def __new__(
+        self,
+        value,
+        allow_limited_range: bool = True,
+        allow_negative_index: bool = True,
+    ):
         return super().__new__(self, value)
 
-    def __init__(self, value, allow_limited_range: bool = True):
+    def __init__(
+        self,
+        value,
+        allow_limited_range: bool = True,
+        allow_negative_index: bool = True,
+    ):
         super().__init__()
         self.allow_limited_range = allow_limited_range
+        self.allow_negative_index = allow_negative_index
 
         # Determine the range type
         self.is_digit = False
@@ -88,6 +100,11 @@ class PathIndexRange(str):
 
         self._validate()
 
+    @property
+    def is_counted_from_back(self):
+        """Is the range counted from the back end"""
+        return self.is_digit and int(self) < 0
+
     @property
     def start(self) -> Union[int, None]:
         """First index in range"""
@@ -105,8 +122,17 @@ class PathIndexRange(str):
                 return int(start)
         return None
 
+    def _validate_digit(self):
+        if not self.allow_negative_index and int(self) < 0:
+            raise HubitError(f"Index '{self}' should be positive")
+
     def _validate_limited_range(self):
 
+        if not self.allow_limited_range:
+            raise HubitError(
+                f"Limited range is not allowed here. Found range '{self}'."
+            )
+
         start, end = self.split(self.wildcard_chr)
 
         if not (start == "" or is_digit(start)):
@@ -135,15 +161,8 @@ class PathIndexRange(str):
             )
 
     def _validate(self):
-
-        if self.is_limited_range and not self.allow_limited_range:
-            raise HubitError(
-                f"Limited range is not allowed here. Found range '{self}'."
-            )
-
         if self.is_digit:
-            if int(self) < 0:
-                raise HubitError(f"Index '{self}' should be positive")
+            self._validate_digit()
         elif self.is_limited_range:
             self._validate_limited_range()
 
@@ -246,20 +265,24 @@ class PathIndexRange(str):
             return str(idx) == self
         elif self.is_full_range:
             return True
+        elif self.is_empty:
+            return True
         else:
-            raise HubitError(f"Unknown range {self}")
+            raise HubitError(f"Unknown range '{self}'")
 
     def intersects(
         self, other: PathIndexRange, allow_self_empty: bool = False
     ) -> Optional[bool]:
         """The intersection between the two ranges
-             self               D  L  F  E
+
+        Scenarios
+
+             self               D  L  F  Empty
         other
                   Digit (D)     2  4  1  6
                   Limited (L)   3  5  1  6
                   Full (F)      1  1  1  6
         """
-
         if self.is_full_range or other.is_full_range:
             # Scenario 1: At least one range covers all indices so
             # there must be an an intersection
@@ -309,8 +332,7 @@ class QueryIndexSpecifier(_IndexSpecifier):
     """
     Index specifiers for [`HubitQueryPath`][hubit.config.HubitQueryPath].
     Currently, index specifiers should be either a positive integer or
-    the character `:`. General slicing and negative indices is not
-    supported.
+    the character `:`. General slicing is not supported.
     """
 
     @property
@@ -326,7 +348,7 @@ class QueryIndexSpecifier(_IndexSpecifier):
         return 0
 
 
-class ModelIndexSpecifier(str):
+class ModelIndexSpecifier(_IndexSpecifier):
     """
     Index specifiers for [`HubitModelPath`][hubit.config.HubitModelPath].
     A model path index specifier is composed of three parts namely the
@@ -375,12 +397,17 @@ class ModelIndexSpecifier(str):
     # Characters a-z, A-Z, _ and digits are allowed
     regex_allowed_identifier = r"^[a-zA-Z_0-9]+$"
 
-    def __new__(self, value, allow_limited_range: bool = True):
+    def __new__(
+        self, value, allow_limited_range: bool = True, allow_negative_index: bool = True
+    ):
         return super().__new__(self, value)
 
-    def __init__(self, value, allow_limited_range: bool = True):
+    def __init__(
+        self, value, allow_limited_range: bool = True, allow_negative_index: bool = True
+    ):
         super().__init__()
         self.allow_limited_range = allow_limited_range
+        self.allow_negative_index = allow_negative_index
 
     def validate(self):
         assert (
@@ -424,6 +451,7 @@ class ModelIndexSpecifier(str):
             PathIndexRange(
                 self.split(self.ref_chr)[0],
                 allow_limited_range=self.allow_limited_range,
+                allow_negative_index=self.allow_negative_index,
             )
             if self.ref_chr in self
             else PathIndexRange("")
@@ -459,6 +487,13 @@ class ModelIndexSpecifier(str):
         return cls(f"{_idx_range}{identifier}{_offset}")
 
 
+Path = Union["HubitModelPath", "HubitQueryPath"]
+
+# upper bound types
+# Path = TypeVar("Path", bound="_HubitPath") # gives issues with init LengthTree._prune_path
+IndexSpecifier = TypeVar("IndexSpecifier", bound="_IndexSpecifier")
+
+
 class _HubitPath(str):
     wildcard_chr = PathIndexRange.wildcard_chr
 
@@ -467,6 +502,10 @@ class _HubitPath(str):
     # Any sequence of characters in square braces including the braces
     regex_braces = r"\[.*?\]"
 
+    @property
+    def index_context(self):
+        raise NotImplementedError
+
     @staticmethod
     def as_internal(path: Any) -> str:
         """Convert path using braces [IDX] to internal
@@ -519,6 +558,12 @@ class _HubitPath(str):
     def components(self):
         return self.as_internal(self).split(".")
 
+    def new_with_index(self, idxspec: IndexSpecifier, index: str) -> Path:
+        """
+        Replace the first index specifier (idxspec) with the index
+        """
+        return cast(Path, self.__class__(self.replace(f"[{idxspec}]", f"[{index}]", 1)))
+
     def set_indices(self, indices: Sequence[str], mode: int = 0) -> _HubitPath:
         """Replace the index specifiers on the path with location indices
 
@@ -590,7 +635,7 @@ class _HubitPath(str):
 
     def has_slice_range(self):
         """Check if path has a slice that is a range"""
-        return self.wildcard_chr in self.ranges()
+        return self.wildcard_chr in self
 
     def ranges(self) -> List[PathIndexRange]:
         """Get the ranges from the path i.e. the part of all square braces preceding the @.
@@ -610,10 +655,6 @@ class _HubitPath(str):
         return re.split(f"{_HubitPath.regex_braces}.|{_HubitPath.regex_braces}", self)
 
 
-# upper bound type
-Path = TypeVar("Path", bound=_HubitPath)
-
-
 class HubitQueryPath(_HubitPath):
     """
     Reference a field in the results data. The syntax follows general
@@ -668,21 +709,19 @@ class HubitQueryPath(_HubitPath):
     #         _path = _path.replace(idx_spec, index, 1)
     #     return self.__class__(_path)
 
-    def check_path_match(self, model_path: HubitModelPath) -> bool:
-        """Check if the query matches the model path from the
-        model bindings
-
-        Args:
-            model_path: A model path (provider)
-
-        Returns:
-            bool: True if the query matches the model path
+    def field_names_match(self, model_path: HubitModelPath) -> bool:
+        """Check if the field names in query match the field names
+        in the model path
         """
-        q_fields = self.remove_braces()
-        m_fields = model_path.remove_braces()
-        if q_fields != m_fields:
+        if self.remove_braces() != model_path.remove_braces():
             return False
+        else:
+            return True
 
+    def index_specifiers_match(self, model_path: HubitModelPath) -> bool:
+        """Check if the index specifiers in query intersect the index
+        specifiers in the model path
+        """
         q_specifiers = self.get_index_specifiers()
         m_specifiers = model_path.get_index_specifiers()
         if len(q_specifiers) != len(m_specifiers):
@@ -693,20 +732,40 @@ class HubitQueryPath(_HubitPath):
                 return False
         return True
 
+    def path_match(self, model_path: HubitModelPath) -> bool:
+        """Check if the query matches the model path from the
+        model bindings
+
+        Args:
+            model_path: A model path (provider)
+            check_intersection: Should the intersection between the index specifiers be checked
+
+        Returns:
+            bool: True if the query matches the model path
+        """
+        if not self.field_names_match(model_path):
+            return False
+        else:
+            return self.index_specifiers_match(model_path)
+
     def idxs_for_matches(
-        self,
-        mpaths: List[HubitModelPath],
+        self, mpaths: List[HubitModelPath], check_intersection: bool = True
     ) -> List[int]:
         """
         Returns indices in the sequence of provider strings that match the
         structure of the query string
         """
-        return [idx for idx, mpath in enumerate(mpaths) if self.check_path_match(mpath)]
+        feval = self.path_match if check_intersection else self.field_names_match
+        return [idx for idx, mpath in enumerate(mpaths) if feval(mpath)]
 
     def get_index_specifiers(self) -> List[QueryIndexSpecifier]:
         items = re.findall(_HubitPath.regex_idx_spec, self)
         return [QueryIndexSpecifier(item) for item in items]
 
+    @property
+    def has_negative_indices(self):
+        return any(range_.is_counted_from_back for range_ in self.ranges())
+
 
 class _HubitQueryDepthPath(HubitQueryPath):
     """
@@ -836,12 +895,13 @@ class HubitModelPath(_HubitPath):
 
     """
 
-    def __new__(self, value, allow_limited_range: bool = True):
+    def __new__(self, value: str = "", allow_limited_range: bool = True):
         return super().__new__(self, value)
 
-    def __init__(self, value, allow_limited_range: bool = True):
+    def __init__(self, value: str = "", allow_limited_range: bool = True):
         super().__init__()
         self.allow_limited_range = allow_limited_range
+        self.allow_negative_index = True
 
     def _validate_index_specifiers(self):
         for idx_spec in self.get_index_specifiers():
@@ -856,7 +916,12 @@ class HubitModelPath(_HubitPath):
 
     def get_index_specifiers(self) -> List[ModelIndexSpecifier]:
         items = re.findall(_HubitPath.regex_idx_spec, self)
-        return [ModelIndexSpecifier(item, self.allow_limited_range) for item in items]
+        return [
+            ModelIndexSpecifier(
+                item, self.allow_limited_range, self.allow_negative_index
+            )
+            for item in items
+        ]
 
     def get_index_identifiers(self) -> List[str]:
         """Get the index identifiers from the path i.e. the
@@ -902,7 +967,8 @@ class HubitModelPath(_HubitPath):
     def as_include_pattern(self):
         return self.remove_braces()
 
-    def get_idx_context(self):
+    @property
+    def index_context(self):
         """
         Get the index context of a path
         """
@@ -944,13 +1010,20 @@ class HubitBinding:
         """
 
         # Do not allow limited index ranges in binding paths
-        return cls(
-            name=cfg["name"],
-            path=HubitModelPath(
-                cfg["path"],
-                allow_limited_range=False,
-            ),
-        ).validate()
+        try:
+            return cls(
+                name=cfg["name"],
+                path=HubitModelPath(
+                    cfg["path"],
+                    allow_limited_range=False,
+                ),
+            ).validate()
+        except HubitError as err:
+            name = cfg["name"]
+            path = cfg["path"]
+            raise HubitError(
+                f"Error while processing binding with name '{name}' and path '{path}': {str(err)}"
+            )
 
 
 @dataclass
@@ -1146,6 +1219,18 @@ class HubitModelComponent:
     def include_patterns(self):
         return [binding.path.as_include_pattern() for binding in self.consumes_input]
 
+    @property
+    def provides_results_paths(self) -> List[HubitModelPath]:
+        return [binding.path for binding in self.provides_results]
+
+    @property
+    def consumes_input_paths(self) -> List[HubitModelPath]:
+        return [binding.path for binding in self.consumes_input]
+
+    @property
+    def consumes_results_paths(self) -> List[HubitModelPath]:
+        return [binding.path for binding in self.consumes_results]
+
 
 @dataclass
 class HubitModelConfig:
@@ -1254,6 +1339,27 @@ class HubitModelConfig:
 
         return cls(components=components, _base_path=base_path).validate()
 
+    def provides(self) -> Tuple[List[str], List[HubitModelPath]]:
+        """Returns two lists with
+        1) component ID
+        2) Binding path for attributes provided with scope set
+        """
+        cmp_ids, paths_provided, scopes = zip(
+            *[
+                (cmp.id, binding.path, cmp.index_scope)
+                for cmp in self.components
+                for binding in cmp.provides_results
+            ]
+        )
+
+        # Set the scope to check if provided paths are unique
+        paths_provided = [
+            path.set_range_for_idxid(scope)
+            for path, scope in zip(paths_provided, scopes)
+        ]
+
+        return cmp_ids, paths_provided
+
 
 @dataclass
 class Query:
