@@ -14,29 +14,24 @@ import pickle
 from dataclasses import dataclass, field, fields
 from typing import (
     Any,
-    Callable,
     List,
     Sequence,
     Tuple,
     Dict,
     Optional,
-    cast,
     TYPE_CHECKING,
 )
-from multiprocessing.managers import SyncManager
-from multiprocessing import Manager
 import logging
 import os
 import time
 import copy
 import itertools
 from multiprocessing import Pool
-from threading import Thread, Event
+import warnings
 
-from .qrun import _QueryRunner
+from .qrun import query_runner_factory
 from .tree import LengthTree, _QueryExpansion, tree_for_idxcontext
 from .config import FlatData, HubitModelConfig, Query, PathIndexRange, HubitQueryPath
-from .utils import set_element
 from .render import get_dot
 
 from .errors import (
@@ -49,7 +44,7 @@ from .errors import (
 
 if TYPE_CHECKING:
     from .config import HubitBinding, HubitModelComponent, HubitModelPath
-
+    from .qrun import _QueryRunner
 
 _CACHE_DIR = ".hubit_cache"
 _HUBIT_DIR = os.path.dirname(os.path.realpath(__file__))
@@ -71,7 +66,6 @@ def _get(
     flat_input,
     flat_results: Optional[FlatData] = None,
     dryrun: bool = False,
-    expand_iloc: bool = False,
 ):
     """
     With the 'queryrunner' object deploy the paths
@@ -82,80 +76,21 @@ def _get(
     If dryrun=True the workers will generate dummy results. Usefull
     to validate s query.
     """
-    if flat_results is None:
-        _flat_results = FlatData()
-    else:
-        _flat_results = flat_results
-    # Reset book keeping data
-    queryrunner.workers = []
-    queryrunner.workers_working = []
-    queryrunner.workers_completed = []
-    queryrunner.worker_for_id = {}
-    queryrunner.observers_for_query = {}
 
-    extracted_input: Dict[str, Any] = {}
+    queries_exp, err, status = queryrunner.run(query, flat_input, flat_results, dryrun)
 
-    # Expand the query for each path
-    queries_exp = [queryrunner.model._expand_query(qpath) for qpath in query.paths]
-    # Make flat list of expanded queries
-    _queries = [qpath for qexp in queries_exp for qpath in qexp.flat_expanded_paths()]
+    if err is None:
+        # TODO: compression call belongs on model (like expand)
+        response = queryrunner.model._collect_results(
+            queryrunner.flat_results, queries_exp
+        )
 
-    for qexp in queries_exp:
-        logging.debug(f"Expanded query {qexp}")
-
-    # Start thread that periodically checks whether we are finished or not
-    shutdown_event = Event()
-    watcher = Thread(
-        target=queryrunner._watcher, args=(_queries, _flat_results, shutdown_event)
-    )
-    watcher.daemon = True
-
-    # remeber to send SIGTERM for processes
-    # https://stackoverflow.com/questions/11436502/closing-all-threads-with-a-keyboard-interrupt
-    the_err = None
-    try:
-        watcher.start()
-        if queryrunner.use_multi_processing:
-            with Manager() as manager:
-                # mypy complains although the type seems to be SyncManager as expected
-                manager = cast(SyncManager, manager)
-                queryrunner.spawn_workers(
-                    _queries,
-                    extracted_input,
-                    _flat_results,
-                    flat_input,
-                    manager,
-                    dryrun=dryrun,
-                )
-                watcher.join()
-        else:
-            queryrunner.spawn_workers(
-                _queries,
-                extracted_input,
-                _flat_results,
-                flat_input,
-                dryrun=dryrun,
-            )
-            watcher.join()
-
-    except (Exception, KeyboardInterrupt) as err:
-        the_err = err
-        shutdown_event.set()
-
-    # Join workers
-    queryrunner._join_workers()
-
-    if the_err is None:
-        response = {query: _flat_results[query] for query in _queries}
-
-        if not expand_iloc:
-            # TODO: compression call belongs on model (like expand)
-            response = queryrunner.model._compress_response(response, queries_exp)
-
-        return response, _flat_results
+        return response, queryrunner.flat_results
     else:
+        print("Exited with runner status")
+        print(status)
         # Re-raise if failed
-        raise the_err
+        raise err
 
 
 class HubitModel:
@@ -190,15 +125,9 @@ class HubitModel:
         # Stores length tree. Filled when set_input() is called
         self.tree_for_idxcontext: Dict[str, LengthTree] = {}
 
-        # Stores trees for query
-        self._tree_for_qpath: Dict[str, LengthTree] = {}
-
         # Stores normalized query paths
         self._normqpath_for_qpath: Dict[str, str] = {}
 
-        # Store the model path that matches the query
-        self._modelpath_for_querypath: Dict[str, str] = {}
-
         self.name = name
         self.odir = os.path.normpath(os.path.join(model_cfg.base_path, output_path))
         self.inputdata: Dict[str, Any] = {}
@@ -370,6 +299,20 @@ class HubitModel:
         """
         Get model results as [`FlatData`][hubit.config.FlatData]
 
+        Returns:
+            Results for the model instance.
+        """
+        warnings.warn(
+            "The method 'get_results' on 'HubitModel' is deprecated. Use the 'results' property instead.",
+            DeprecationWarning,
+        )
+        return self.flat_results
+
+    @property
+    def results(self) -> FlatData:
+        """
+        Get model results as [`FlatData`][hubit.config.FlatData]
+
         Returns:
             Results for the model instance.
         """
@@ -418,9 +361,9 @@ class HubitModel:
 
         _query = Query.from_paths(query)
 
-        # Make a query runner
-        self._qrunner = _QueryRunner(
-            self, use_multi_processing, self._component_caching
+        # Create a query runner
+        self._qrunner = query_runner_factory(
+            use_multi_processing, self, self._component_caching
         )
 
         if validate:
@@ -509,10 +452,9 @@ class HubitModel:
 
             if skipfun(_flat_input):
                 continue
-            qrun = _QueryRunner(
-                self,
-                use_multi_processing=False,
-                component_caching=self._component_caching,
+            # Never use multiprocessing in the worker pool
+            qrun = query_runner_factory(
+                False, self, component_caching=self._component_caching
             )
             flat_results = FlatData()
             args.append((qrun, _query, _flat_input, flat_results))
@@ -653,12 +595,12 @@ class HubitModel:
             self.model_cfg.component_for_id.values(), self.inputdata
         )
 
-    def _validate_query(self, query: Query, use_multi_processing=False):
+    def _validate_query(self, query: Query, use_multi_processing: bool = False):
         """
         Run the query using a dummy calculation to see that all required
         input and results are available
         """
-        qrunner = _QueryRunner(self, use_multi_processing)
+        qrunner = query_runner_factory(use_multi_processing, self)
         _get(qrunner, query, self.flat_input, dryrun=True)
         return qrunner.workers
 
@@ -672,25 +614,19 @@ class HubitModel:
                 else:
                     raise HubitModelValidationError(binding.path, fname, fname_for_path)
 
-    def _cmpids_for_query(self, qpath: HubitQueryPath) -> List[str]:
+    def _cmpids_for_query(
+        self, qpath: HubitQueryPath, check_intersection: bool = True
+    ) -> List[str]:
         """
         Find IDs of components that can respond to the "query".
         """
-        # TODO: Next two lines should only be executed once in init (speed)
-        itempairs = [
-            (cmp.id, binding.path, cmp.index_scope)
-            for cmp in self.model_cfg.components
-            for binding in cmp.provides_results
-        ]
-        cmp_ids, paths_provided, scopes = zip(*itempairs)
+        # TODO: Next line should only be executed once in init (speed)
+        cmp_ids, paths_provided = self.model_cfg.provides()
 
-        # Set the scope to check if provided paths are unique
-        _paths_provided = [
-            path.set_range_for_idxid(scope)
-            for path, scope in zip(paths_provided, scopes)
+        return [
+            cmp_ids[idx]
+            for idx in qpath.idxs_for_matches(paths_provided, check_intersection)
         ]
-        idxs = qpath.idxs_for_matches(_paths_provided)
-        return [cmp_ids[idx] for idx in idxs]
 
     def component_for_id(self, compid: str) -> HubitModelComponent:
         return self.model_cfg.component_for_id[compid]
@@ -723,30 +659,31 @@ class HubitModel:
     def component_for_qpath(self, path: HubitQueryPath) -> HubitModelComponent:
         return self.component_for_id(self._cmpid_for_query(path))
 
-    def _mpaths_for_qpath(self, qpath: HubitQueryPath) -> List[HubitModelPath]:
+    def mpaths_for_qpath_fields_only(
+        self, qpath: HubitQueryPath
+    ) -> Tuple[List[HubitModelPath], List[str]]:
+        """Find model paths that match the query. The match
+        is evaluated only based on field names
         """
-        Returns the model paths (with the index scope inserted)
-        that match the query.
-        """
-        # Find component that provides queried result
-        cmp_ids = self._cmpids_for_query(qpath)
-        # Find component
-        paths = []
-        scopes = []
-        for cmp_id in cmp_ids:
-            cmp = self.model_cfg.component_for_id[cmp_id]
-            # Find index in list of binding paths that match query path
-            idxs = qpath.idxs_for_matches(
-                [binding.path for binding in cmp.provides_results]
-            )
-            paths.append(cmp.provides_results[idxs[0]].path)
-            scopes.append(cmp.index_scope)
+        check_intersection = False
 
-        # Set the index scope
-        mpaths = [
-            mpath.set_range_for_idxid(scope) for mpath, scope in zip(paths, scopes)
-        ]
-        return mpaths
+        # Find component that provides queried result
+        cmp_ids = self._cmpids_for_query(qpath, check_intersection=check_intersection)
+
+        # Find component
+        mpaths = []
+        for cmp_id in cmp_ids:
+            # Get the provided paths
+            binding_paths = self.model_cfg.component_for_id[
+                cmp_id
+            ].provides_results_paths
+            # Find index in list of binding paths that match query path (there can
+            # be only one match which should be validated elsewhere)
+            idx = qpath.idxs_for_matches(
+                binding_paths, check_intersection=check_intersection
+            )[0]
+            mpaths.append(binding_paths[idx])
+        return mpaths, cmp_ids
 
     def _expand_query(
         self, qpath: HubitQueryPath, store: bool = True
@@ -759,88 +696,36 @@ class HubitModel:
         store: If True some intermediate results will be saved on the
             instance for later use.
 
-        TODO: NEgative indices... prune_tree requires real indices but normalize
-        path require all IDX_WILDCARDs be expanded to get the context
-
-        # TODO: Save pruned trees so the worker need not prune top level trees again
-        # TODO: save component so we dont have to find top level components again
+        TODO: Save pruned trees so the worker need not prune top level trees again
+        TODO: save component so we dont have to find top level components again
         """
-        # Get all model paths that match the query
-        mpaths = self._mpaths_for_qpath(qpath)
 
-        # Prepare query expansion object
-        qexp = _QueryExpansion(qpath, mpaths)
+        # Find mpaths and component IDs
+        mpaths, cmp_ids = self.mpaths_for_qpath_fields_only(qpath)
+
+        # Find components corresponding to the IDs
+        cmps = [self.model_cfg.component_for_id[cmp_id] for cmp_id in cmp_ids]
 
         # Get the tree that corresponds to the (one) index context
-        tree = self.tree_for_idxcontext[qexp.idx_context]
+        index_context = _QueryExpansion.get_index_context(qpath, mpaths)
+        tree = self.tree_for_idxcontext[index_context]
 
-        # Validate that tree and expantion are consistent
-        qexp.validate_tree(tree)
+        # Create and return expansion object
+        return _QueryExpansion(qpath, mpaths, tree, cmps)
 
-        # First store unpruned tree
-        if store:
-            self._tree_for_qpath[qpath] = tree
-
-        for decomposed_qpath, _mpath in zip(qexp.decomposed_paths, mpaths):
-            pruned_tree = tree.prune_from_path(
-                decomposed_qpath,
-                inplace=False,
-            )
-
-            if store:
-                # Store pruned tree
-                self._tree_for_qpath[decomposed_qpath] = pruned_tree
-                self._modelpath_for_querypath[decomposed_qpath] = _mpath
-
-            # Expand the path
-            expanded_paths = pruned_tree.expand_path(decomposed_qpath, flat=True)
-
-            qexp.update_expanded_paths(decomposed_qpath, expanded_paths)
-
-        return qexp
-
-    def _compress_response(self, response, queries_expanded: List[_QueryExpansion]):
+    def _collect_results(
+        self, flat_results: FlatData, query_expansions: List[_QueryExpansion]
+    ):
         """
         Compress the response to reflect queries with index wildcards.
         So if the query has the structure list1[:].list2[:] and is
-        rectangular with N1 (2) elements in list1 and N2 (3) elements
+        rectangular with 2 elements in list1 and 3 elements
         in list2 the compressed response will be a nested list like
         [[00, 01, 02], [10, 11, 12]]
         """
-        _response = {}
-        for qexp in queries_expanded:
-            if not qexp.is_expanded():
-                # Path not expanded so no need to compress
-                _response[qexp.path] = response[qexp.path]
-            else:
-                # Get the index IDs from the original query
-                idxids = qexp.path.get_index_specifiers()
-                tree = self._tree_for_qpath[qexp.path]
-                # TODO: Can we prune earlier on?
-                # inplace = False to leave the model state unchanged.
-                # This is important for successive get requests
-                values_decomp = tree.prune_from_path(
-                    qexp.path, inplace=False
-                ).none_like()
-                # Extract iloc indices for each query in the expanded query
-                for expanded_paths in qexp.expanded_paths_for_decomposed_path.values():
-                    for path in expanded_paths:
-                        ranges = path.ranges()
-                        # Only keep ilocs that come from an expansion... otherwise
-                        # the dimensions of "values" do no match
-                        ranges = [
-                            range
-                            for range, idxid in zip(ranges, idxids)
-                            if idxid == IDX_WILDCARD
-                        ]
-                        values_decomp = set_element(
-                            values_decomp,
-                            response[path],
-                            [int(range) for range in ranges],
-                        )
-                _response[qexp.path] = values_decomp
-
-        return _response
+        return {
+            qexp.path: qexp.collect_results(flat_results) for qexp in query_expansions
+        }
 
 
 def now():
