@@ -12,15 +12,22 @@ from typing import (
     Union,
     Optional,
     TYPE_CHECKING,
+    cast,
 )
 from .errors import HubitIndexError, HubitError, HubitModelQueryError
-from .config import _HubitPath, HubitModelPath, HubitQueryPath, PathIndexRange, Path
-from .utils import is_digit, get_from_datadict, split_items, traverse
+from .config import (
+    FlatData,
+    HubitModelPath,
+    HubitQueryPath,
+    PathIndexRange,
+    Path,
+)
+from .utils import is_digit, get_from_datadict, split_items, traverse, set_element
 
 if TYPE_CHECKING:
     from .config import HubitModelComponent
 
-IDX_WILDCARD = PathIndexRange.wildcard_chr
+Node = Union["LeafNode", "LengthNode"]
 
 
 class LeafNode:
@@ -34,9 +41,17 @@ class LeafNode:
         self.level = 0
         self.parent = None
 
+    @property
+    def children(self):
+        """A leaf has no children"""
+        raise NotImplementedError
+
     def remove_decendants(self):
         pass
 
+    def __str__(self):
+        return f'LeafNode(nchildren=NA, index={self.index}, has parent={"Yes" if self.parent else "No"}, is_constrained=NA)'
+
 
 class LengthNode:
     def __init__(self, nchildren: int):
@@ -49,9 +64,8 @@ class LengthNode:
         """
         self.level = 0
         # Assume bottom level
-        self.children: Sequence[Union[LengthNode, LeafNode]] = [
-            LeafNode(idx) for idx in range(nchildren)
-        ]
+        self._nchildren_org = nchildren
+        self.children: Sequence[Node] = [LeafNode(idx) for idx in range(nchildren)]
         self._set_child_for_idx()
 
         # Assume top level (parent = None)
@@ -67,11 +81,39 @@ class LengthNode:
     def _set_child_for_idx(self):
         self._child_for_idx = {child.index: child for child in self.children}
 
+    def child(self, index: int) -> Union[LengthNode, LeafNode]:
+        """Get the child corresponding to the specified index"""
+        # TODO: should this be improved?
+        try:
+            return {child.index: child for child in self.children}[
+                self.normalize_child_index(index)
+            ]
+        except KeyError:
+            idxs = [str(child.index) for child in self.children]
+            raise HubitIndexError(
+                f"No child with index {index} ({self.normalize_child_index(index)}) on node."
+                f"Available indices are: {', '.join(idxs)}"
+            )
+
+    def normalize_child_index(self, index: int):
+        # TODO: think about this for a while...
+        if index < 0:
+            norm_index = self._nchildren_org + index
+            if norm_index < 0:
+                raise HubitIndexError(
+                    f"Index {index} was normalized to {norm_index} for node "
+                    f"that had {self._nchildren_org} children (now {self.nchildren()})."
+                )
+            return norm_index
+        else:
+            return index
+
     def nchildren(self) -> int:
         return len(self.children)
 
     def set_children(self, children: Sequence[Union[LengthNode, LeafNode]]):
         self.children = list(children)
+        self._nchildren_org = len(self.children)
         for idx, child in enumerate(self.children):
             child.parent = self
             child.level = self.level + 1
@@ -125,12 +167,13 @@ class DummyLengthTree:
     def __init__(self, *args, **kwargs):
         self.level_names = []
 
-    def get_idx_context(self):
+    @property
+    def index_context(self):
         return "-".join(self.level_names)
 
     def prune_from_path(
         self,
-        _path: _HubitPath = _HubitPath(),
+        _path: Path,
         inplace: bool = True,
     ) -> DummyLengthTree:
         return self if inplace else copy.deepcopy(self)
@@ -151,14 +194,15 @@ class DummyLengthTree:
         """Don't change the tree"""
         return self if inplace else copy.deepcopy(self)
 
-    def expand_path(
-        self, path: HubitModelPath, *args, **kwargs
-    ) -> List[HubitModelPath]:
-        return [path]
+    def expand_path(self, path: Path, *args, **kwargs) -> List[HubitQueryPath]:
+        return [HubitQueryPath(path)]
 
     def none_like(self):
         return None
 
+    def is_path_described(self, path: Path) -> bool:
+        return True
+
 
 class LengthTree:
     """Stores length information for multi-dimensional and non-rectangular
@@ -173,17 +217,22 @@ class LengthTree:
             nodes (List[LengthNode]): Connected length nodes
             level_names (List[str]): Name of the levels specified on the nodes
         """
-        self.nlevels = len(level_names)
         self.level_names: Sequence[str] = level_names
 
-        self.nodes_for_level: Sequence[List[LengthNode]] = [
-            [] for _ in range(self.nlevels)
-        ]
+        # Set to pruning path when pruned else None
+        self._prune_path: Union[None, Path] = None
+
+        self.nodes_for_level: Sequence[List[LengthNode]] = [[] for _ in level_names]
         for node in nodes:
             node.tree = self
             self.nodes_for_level[node.level].append(node)
 
-    def get_idx_context(self):
+    @property
+    def nlevels(self):
+        return len(self.nodes_for_level)
+
+    @property
+    def index_context(self):
         return "-".join(self.level_names)
 
     def clip_at_level(self, level_name: str, inplace: bool = True) -> LengthTree:
@@ -204,7 +253,7 @@ class LengthTree:
         obj.nodes_for_level = obj.nodes_for_level[: level_idx + 1]
         return obj
 
-    def prune_from_path(self, path: _HubitPath, inplace: bool = True) -> LengthTree:
+    def prune_from_path(self, path: Path, inplace: bool = True) -> LengthTree:
         """Prune the length tree based on a path where zero
         to all indices are already specified.
 
@@ -216,6 +265,7 @@ class LengthTree:
             LengthTree: Pruned tree. If inplace=True the instance itself is also pruned.
         """
         obj = self if inplace else copy.deepcopy(self)
+        obj._prune_path = path
         ranges = path.ranges()
         if all([range.is_full_range for range in ranges]):
             return obj
@@ -256,9 +306,10 @@ class LengthTree:
             # Keep child corresponding to idx remove the rest
             idx_others = [child.index for child in node.children]
             node.is_constrained = True
-            if idx_value in idx_others:
+            idx_value_norm = node.normalize_child_index(idx_value)
+            if idx_value_norm in idx_others:
                 # idx can be provided
-                idx_others.remove(idx_value)
+                idx_others.remove(idx_value_norm)
                 for idx_other in reversed(idx_others):
                     node.pop_child_for_idx(idx_other)
 
@@ -270,7 +321,7 @@ class LengthTree:
                 node.remove()
             except HubitIndexError:
                 raise HubitIndexError(
-                    f"Cannot find index {idx_value} "
+                    f"Cannot find index {idx_value} ({idx_value_norm}) "
                     f"for index ID {obj.level_names[level_idx]} "
                     f"in tree "
                     f"{self}"
@@ -477,33 +528,17 @@ class LengthTree:
                 node = node._child_for_idx[node_idx]
         return node
 
-    def normalize_path(self, qpath: HubitQueryPath) -> HubitQueryPath:
-        """Handle negative indices
-        As stated in "test_normalize_path2" the normalization in general depends
-        on the context
-        """
-
-        idxids = qpath.get_index_specifiers()
-        _path = copy.copy(qpath)
-
-        for idx_level, idxid in enumerate(idxids):
-            if is_digit(idxid) and int(idxid) < 0:
-                # Get index context i.e. indices prior to current level
-                _idx_context = [
-                    int(idx) for idx in _path.get_index_specifiers()[:idx_level]
-                ]
-                node = self._node_for_idxs(_idx_context)
-                _path = HubitQueryPath(
-                    _path.replace(idxid, str(node.nchildren() + int(idxid)), 1)
-                )
-        return _path
+    def _is_pruned_from(self, path: Path):
+        return self._prune_path == path
 
     def expand_path(
         self,
         path: Path,
         flat: bool = False,
     ) -> List[HubitQueryPath]:
-        """Expand model path with wildcard based on tree
+        """Expand model path with wildcard based on tree. If path is fixed at some index the
+        tree must be pruned according to the specified path if flat = False. This is not
+        checked.
 
         Example for a query path:
             list[:].some_attr.numbers ->
@@ -519,40 +554,33 @@ class LengthTree:
             defined by the tree if flat = False. Otherwise a
             flat list.
         """
+        if not self._is_pruned_from(path):
+            raise HubitError(
+                f"Tree should be pruned using path '{path}'. Tree pruned with: '{self._prune_path}'"
+            )
+
         # Get the content of the braces
         idxspecs = path.get_index_specifiers()
-        ranges = path.ranges()
 
-        # Manipulate as strings
-        paths = [str(path)]
-        for idx_level, (idxspec, range) in enumerate(zip(idxspecs, ranges)):
+        # Top level always has one node
+        paths = [path]
+
+        # Loop over levels in sequence stating from the top (left)
+        for idx_level, idxspec in enumerate(idxspecs):
             nodes = self.nodes_for_level[idx_level]
-            paths_current_level = []
+            # After processing the number of paths (on the current level) there will
+            # always be an equal number of paths and children. Therefore, before the
+            # processing (here), the number of nodes and paths are
+            # equal (in a LengthTree #nodes next level = #children current level).
+            paths_current_level: List[Path] = []
             for _path, node in zip(paths, nodes):
-                if range.is_digit:
-                    # slice is digit so replace index specifier with that digit
-                    paths_current_level.append(_path.replace(idxspec, range))
-                elif range.is_full_range or range.is_empty:
-                    # slice is wildcard or not specified so expand from node children
-                    paths_current_level.extend(
-                        [
-                            _path.replace(idxspec, str(child.index), 1)
-                            for child in node.children
-                        ]
-                    )
-                elif range.is_limited_range:
-                    # Loop over children to see who are included in the range
-                    paths_current_level.extend(
-                        [
-                            _path.replace(idxspec, str(child.index), 1)
-                            for child in node.children
-                            if range.contains_index(child.index)
-                        ]
-                    )
-                else:
-                    raise HubitError(
-                        f"Unknown index range '{range}' for path '{path}' of type '{type(path)}'."
-                    )
+                # Assumes pruned tree
+                paths_current_level.extend(
+                    [
+                        _path.new_with_index(idxspec, str(child.index))
+                        for child in node.children
+                    ]
+                )
             paths = copy.deepcopy(paths_current_level)
 
         # Cast strings as paths
@@ -571,6 +599,12 @@ class LengthTree:
         """Number of leaves in the tree"""
         return sum(self.number_of_children(-1))
 
+    def children_at_level(self, idx_level: int) -> List[Node]:
+        """Number of children for each node at the specified level"""
+        return [
+            child for node in self.nodes_for_level[idx_level] for child in node.children
+        ]
+
     def number_of_children(self, idx_level: int) -> List[int]:
         """Number of children for each node at the specified level"""
         return [node.nchildren() for node in self.nodes_for_level[idx_level]]
@@ -590,6 +624,64 @@ class LengthTree:
         # like all leaves in the tree
         return self.reshape([None for _ in range(self.number_of_leaves())])
 
+    def is_path_described(self, path: Path) -> bool:
+        """Check if the specified model path is described by the tree
+
+        On each level the path index range must contain at least one child index. Example:
+        For the tree
+          level=0 (IDX_SITE), nodes=1 parents=0, children=1, children are leaves=[False], child idxs=[[0]]
+          level=1 (IDX_LINE), nodes=1 parents=1, children=1, children are leaves=[False], child idxs=[[0]]
+          level=2 (IDX_TANK), nodes=1 parents=1, children=3, children are leaves=[True], child idxs=[[0, 1, 2]]
+        The following path match
+          prod_sites[IDX_SITE].prod_lines[IDX_LINE].tanks[0@IDX_TANK].Q_yield -> OK
+          prod_sites[IDX_SITE].prod_lines[IDX_LINE].tanks[1@IDX_TANK].Q_yield -> OK
+          prod_sites[IDX_SITE].prod_lines[IDX_LINE].tanks[2@IDX_TANK].Q_yield -> OK
+        while the path below does not
+          prod_sites[IDX_SITE].prod_lines[IDX_LINE].tanks[3@IDX_TANK].Q_yield -> Not OK
+        since the index range '3' coming from [3@IDX_TANK] is not a part of the tree
+        """
+
+        # Index contexts must match
+        try:
+            if not self.index_context == path.index_context:
+                return False
+        except NotImplementedError:
+            # query paths have no index context
+            pass
+
+        ranges = path.ranges()
+
+        # The path must have ranges corresponding to the number of levels in the tree
+        if not (len(ranges) == self.nlevels):
+            return False
+
+        def filter_nodes(range_: PathIndexRange, nodes: List[Node]) -> List[Node]:
+            """Only consider children that are described by the index"""
+            return [node for node in nodes if range_.contains_index(node.index)]
+
+        def nodes_next(nodes: List[Node]) -> List[Node]:
+            """Get children for the LengthNode"""
+            return [child for node in nodes for child in node.children]
+
+        nodes = self.children_at_level(0)
+
+        # Loop over levels excluding the last which has LeafNode (with no children)
+        for range_ in ranges[:-1]:
+            # Build list of children matching the "description"
+            nodes = filter_nodes(range_, nodes)
+            if len(nodes) == 0:
+                return False
+
+            nodes = nodes_next(nodes)
+
+        # Handle LeafNodes
+        nodes = filter_nodes(ranges[-1], nodes)
+        if len(nodes) == 0:
+            return False
+
+        # If we get to the leaves the path is described
+        return True
+
     def __str__(self):
         lines = ["--------------------", "Tree"]
         for idx, (name, nodes) in enumerate(
@@ -630,99 +722,248 @@ class LengthTree:
 class _QueryExpansion:
     """A Hubit query expansion. A query can be split into multiple queries
 
-    Args:
-        path: A [`HubitQueryPath`][hubit.config.HubitQueryPath] representing the original query.
-        decomposed_paths: If a single component can provide results for `path`, `decomposed_paths`
-            has one element of type [`HubitQueryPath`][hubit.config.HubitQueryPath]. If multiple
-            components match the query individual path contributions are the items in the list.
-        expanded_paths_for_decomposed_path: For each element in `decomposed_paths`
-            these are the expanded paths i.e. dotted paths with real indices not
-            wildcards.
+    path: A [`HubitQueryPath`][hubit.config.HubitQueryPath] representing the original query.
+    decomposed_paths: If a single component can provide results for `path`, `decomposed_paths`
+        has one element of type [`HubitQueryPath`][hubit.config.HubitQueryPath]. If multiple
+        components match the query individual path contributions are the items in the list.
+    exp_paths_for_decomp_path: For each element in `decomposed_paths`
+        these are the expanded paths i.e. dotted paths with real indices not
+        wildcards.
     """
 
-    def __init__(self, path: HubitQueryPath, mpaths: List[HubitModelPath]):
+    def __init__(
+        self,
+        path: HubitQueryPath,
+        mpaths: List[HubitModelPath],
+        tree: LengthTree,
+        cmps: List[HubitModelComponent],
+    ):
         """
         path: the query path
         mpaths: the model paths that match the query
+        cmps: the component for each of the mpaths. If None no mpath filtering is performed
+
+        We don't normalize on initialization  to reduce coupling to model and tree objects
         """
         self.path = path
+        self.tree = tree
+        self.paths_norm = _QueryExpansion._normalize_path(self.path, tree)
 
-        if len(mpaths) > 1 and not path.has_slice_range():
+        # Filter models paths using index specifiers for normalized query path
+
+        # Store in dict with normalized path so decomposition can be carried out in
+        # pairs (qpath_norm, mpaths)
+        index_scopes = [cmp.index_scope for cmp in cmps]
+        mpaths_for_qpath_norm = {
+            qpath_norm: _QueryExpansion._filter_mpaths_for_qpath_index_ranges(
+                qpath_norm,
+                mpaths,
+                index_scopes,
+                self.tree.prune_from_path(qpath_norm, inplace=False),
+            )
+            for qpath_norm in self.paths_norm
+        }
+
+        # Store the flattened version
+        self.mpaths = list(
+            itertools.chain.from_iterable(mpaths_for_qpath_norm.values())
+        )
+
+        if len(self.mpaths) > 1 and not self.path.has_slice_range():
             # Should not be possible to have multiple providers if the query
             # points to a specific path i.e. has no ranges.
             # TODO: This check could be more strict e.g. the wildcard is located where
             # the mpaths vary
             raise HubitModelQueryError(
-                f"More than one component match the query '{path}'. Matching components provide: {mpaths}."
+                f"More than one component match the query '{self.path}'. Matching components provide: {self.mpaths}."
             )
 
-        # Get the index contexts for doing some tests
-        _idx_contexts = {mpath.get_idx_context() for mpath in mpaths}
-        if len(_idx_contexts) > 1:
-            msg = f"Fatal error. Inconsistent providers for query '{path}': {', '.join(mpaths)}"
-            raise HubitModelQueryError(msg)
+        # Get the index contexts for doing some validation
+        self._idx_context = _QueryExpansion.get_index_context(self.path, self.mpaths)
 
-        if len(_idx_contexts) == 0:
-            msg = f"Fatal error. No provider for query path '{path}'."
-            raise HubitModelQueryError(msg)
+        self.decomposed_paths: List[List[HubitQueryPath]] = []
+        for path_norm in self.paths_norm:
+            _decomposed_paths, index_identifiers = _QueryExpansion.decompose_query(
+                path_norm, mpaths_for_qpath_norm[path_norm]
+            )
+            self.decomposed_paths.append(list(set(_decomposed_paths)))
 
-        self.decomposed_paths, index_identifiers = _QueryExpansion.decompose_query(
-            path, mpaths
-        )
-        self.expanded_paths_for_decomposed_path: Dict[
-            HubitQueryPath, List[HubitQueryPath]
-        ] = {}
+        # TODO: Cannot see that I actually use the keys any longer
+        self.exp_paths_for_decomp_path: Dict[HubitQueryPath, List[HubitQueryPath]] = {}
 
+        # Used to validate tree against expansion
+        self.decomposed_idx_identifiers: List[Union[None, str]] = []
         if index_identifiers is None:
-            self.decomposed_idx_identifier = None
+            self.decomposed_idx_identifiers.append(None)
         else:
             # Cannot occur since len(_idx_contexts) is 1
             # if len(set(index_identifiers)) > 1:
             #     msg = f"Fatal error. Inconsistent decomposition for query '{path}': {', '.join(mpaths)}"
             #     raise HubitModelQueryError(msg)
-            self.decomposed_idx_identifier = index_identifiers[0]
+            self.decomposed_idx_identifiers.append(index_identifiers[0])
 
-        self._idx_context = list(_idx_contexts)[0]
+        # Validate that tree and expantion are consistent
+        self._validate_tree()
+        self._set_expanded_paths()
+
+    @staticmethod
+    def _filter_mpaths_for_qpath_index_ranges(
+        qpath: HubitQueryPath,
+        mpaths: List[HubitModelPath],
+        index_scopes: Sequence[Dict[str, PathIndexRange]],
+        pruned_tree: LengthTree,
+    ) -> List[HubitModelPath]:
+        """
+        each path represents a path provided for the corresponding component.
+        mpaths, index_scopes have same length
+        """
+        # Indexes for models paths that match the query path (index considering intersections)
+        # Set the index scope
+        _mpaths = [
+            mpath.set_range_for_idxid(index_scope)
+            for mpath, index_scope in zip(mpaths, index_scopes)
+        ]
+
+        # TODO negative-indices. split out field check
+        idxs = qpath.idxs_for_matches(_mpaths, check_intersection=True)
+
+        # Check that math exists in pruned tree
+        _mpaths = [
+            _mpaths[idx] for idx in idxs if pruned_tree.is_path_described(_mpaths[idx])
+        ]
+        return _mpaths
+
+    @staticmethod
+    def _normalize_path(
+        qpath: HubitQueryPath, tree: LengthTree
+    ) -> List[HubitQueryPath]:
+        """
+        If the query path has negative indices we must normalize the path
+        i.e expand it to get rid of this abstraction.
+        """
+        if qpath.has_negative_indices:
+            # Get index context to find tree and normalizethe  path. The tree
+            # is required since field[:].filed2[-1] may, for non-rectangular data,
+            # for example correspond to
+            # [ field[0].field2[2],
+            #   field[1].field2[4],
+            #   field[2].filed2[1]
+            # ]
+
+            # Even though the model paths have not yet been filtered based on
+            # index ranges the index context should still be unique
+            return tree.prune_from_path(qpath, inplace=False).expand_path(
+                qpath, flat=True
+            )
+        else:
+            return [qpath]
+
+    @staticmethod
+    def get_index_context(qpath: HubitQueryPath, mpaths: List[HubitModelPath]):
+        """
+        Get the index context and do some validation
+        """
+        # Get the index contexts for doing some tests
+        _idx_contexts = {mpath.index_context for mpath in mpaths}
+        if len(_idx_contexts) > 1:
+            msg = f"Fatal error. Inconsistent providers for query '{qpath}': {', '.join(mpaths)}"
+            raise HubitModelQueryError(msg)
+
+        if len(_idx_contexts) == 0:
+            msg = f"Fatal error. No provider for query path '{qpath}'."
+            raise HubitModelQueryError(msg)
+
+        return list(_idx_contexts)[0]
 
     @property
     def idx_context(self):
         """The (one) index context corresponding to the model paths"""
         return self._idx_context
 
-    def update_expanded_paths(
+    def _set_expanded_paths(self):
+        """Set the expanded paths using the tree"""
+        for decomposed_qpaths in self.decomposed_paths:
+            for decomposed_qpath, _mpath in zip(decomposed_qpaths, self.mpaths):
+                pruned_tree = self.tree.prune_from_path(
+                    decomposed_qpath,
+                    inplace=False,
+                )
+
+                # Expand the path
+                expanded_paths = pruned_tree.expand_path(decomposed_qpath, flat=True)
+
+                self._update_expanded_paths(decomposed_qpath, expanded_paths)
+
+    def _update_expanded_paths(
         self, decomposed_path: HubitQueryPath, expanded_paths: List[HubitQueryPath]
     ):
-        self.expanded_paths_for_decomposed_path[decomposed_path] = expanded_paths
+        self.exp_paths_for_decomp_path[decomposed_path] = expanded_paths
 
     def flat_expanded_paths(self):
         """Returns flat list of expanded paths"""
         return [
-            path
-            for paths in self.expanded_paths_for_decomposed_path.values()
-            for path in paths
+            path for paths in self.exp_paths_for_decomp_path.values() for path in paths
         ]
 
+    def normalization_is_simple(self):
+        return len(self.paths_norm) == 1
+
+    def _path_is_normalized(self):
+        return len(self.paths_norm) > 1 or self.path != self.paths_norm[0]
+
     def is_decomposed(self):
-        return len(self.decomposed_paths) > 1
+        return any(len(paths) > 1 for paths in self.decomposed_paths)
 
     def is_expanded(self):
-        if (
-            not self.is_decomposed()
-            and self.path == self.decomposed_paths[0]
-            and len(self.expanded_paths_for_decomposed_path[self.decomposed_paths[0]])
-            == 1
-            and self.path
-            == self.expanded_paths_for_decomposed_path[self.decomposed_paths[0]][0]
-        ):
-            return False
-        else:
+        if self.path.has_slice_range():
             return True
 
+        return False
+
+    def collect_results(self, flat_results: FlatData):
+        """
+        Collect result from the flat data that belong to the query (expansion)
+        """
+        if not self.is_expanded():
+            # Path not expanded so no need to compress
+            # Simple map from single normalized path to the path
+            # e.g cars[-1].price ['cars[2].price'] or cars[2].price ['cars[2].price']
+            return flat_results[self.paths_norm[0]]
+
+        # Get the index IDs from the original query
+        idxids = self.path.get_index_specifiers()
+
+        # TODO: Can we prune earlier on?
+        # inplace = False to leave the tree state unchanged.
+        values = self.tree.prune_from_path(self.path, inplace=False).none_like()
+        # Extract iloc indices for each query in the expanded query
+        for expanded_paths in self.exp_paths_for_decomp_path.values():
+            for path in expanded_paths:
+                ranges = path.ranges()
+                # Only keep ilocs that come from an expansion... otherwise
+                # the dimensions of "values" do no match
+                ranges = [
+                    range
+                    for range, idxid in zip(ranges, idxids)
+                    if idxid == PathIndexRange.wildcard_chr
+                ]
+                values = set_element(
+                    values,
+                    flat_results[path],
+                    [int(range) for range in ranges],
+                )
+
+        return values
+
     @staticmethod
     def decompose_query(
         qpath: HubitQueryPath, mpaths: List[HubitModelPath]
     ) -> Tuple[List[HubitQueryPath], Union[List[str], None]]:
         """
+        Handles the case where more than one provider required to match query.
+        In that case the query is into (decomposed) queries each having a single
+        provider.
+
         If a single component can provide results for `path`, `decomposed_paths`
         has one element of type [`HubitQueryPath`][hubit.config.HubitQueryPath].
         If multiple components are required to provide the query their individual
@@ -775,7 +1016,7 @@ class _QueryExpansion:
 
         return decomposed_qpaths, index_identifiers
 
-    def validate_tree(self, tree: LengthTree):
+    def _validate_tree(self):
         """Validate that we get the expected number of mpaths in the expansion
 
         Raises if tree is invalid.
@@ -784,46 +1025,39 @@ class _QueryExpansion:
         instead of >=.
         """
         # Don't validate if DummyLengthTree since they can never be decomposed
-        if isinstance(tree, DummyLengthTree):
+        if isinstance(self.tree, DummyLengthTree):
             return
 
         # Don't validate tree against expansion if path was not decomposed
         if not self.is_decomposed():
             return
 
-        n_decomposed_paths = len(self.decomposed_paths)
-        tree_level_name = self.decomposed_idx_identifier
+        for decomposed_paths, tree_level_name in zip(
+            self.decomposed_paths, self.decomposed_idx_identifiers
+        ):
 
-        # Find out which level (index) we are at
-        try:
-            idx_level = tree.level_names.index(tree_level_name)
-        except ValueError as err:
-            print(f"ERROR. Level name '{tree_level_name}' not found in tree")
-            print(tree)
-            print(self)
-            raise HubitError("Query expansion error.")
+            # Find out which level (index) we are at
+            try:
+                self.tree.level_names.index(tree_level_name)
+            except ValueError as err:
+                print(f"ERROR. Level name '{tree_level_name}' not found in tree")
+                print(self.tree)
+                print(self)
+                raise HubitError("Query expansion error.")
 
-        # For each node at the level the number of children should be
-        # greater than or equal to the number for paths in the decomposition
-        results = [n >= n_decomposed_paths for n in tree.number_of_children(idx_level)]
-        if not all(results):
-            print(
-                f"ERROR. "
-                f"Too few children at level {idx_level} with "
-                f"name '{tree_level_name}' of tree. "
-                f"Expected at least {n_decomposed_paths} children "
-                "corresponding to the number of decomposed paths.\n"
-            )
-            print(tree)
-            print(self)
-            raise HubitError("Query expansion error.")
+            for path in decomposed_paths:
+                if not self.tree.is_path_described(path):
+                    print(self.tree)
+                    raise HubitError(
+                        f"Query expansion error. Path '{path}' not described by tree."
+                    )
 
     def __str__(self):
         lines = [f"\nQuery\n  {self.path}"]
         lines.append("Decomposition & expansion")
         for decomp_path, expanded_paths in itertools.zip_longest(
             self.decomposed_paths,
-            self.expanded_paths_for_decomposed_path.values(),
+            self.exp_paths_for_decomp_path.values(),
             fillvalue=None,
         ):
             lines.append(f"  {decomp_path} -> {expanded_paths}")
@@ -846,7 +1080,7 @@ def tree_for_idxcontext(
     for component in components:
         for binding in component.consumes_input:
             tree = LengthTree.from_data(binding.path, data)
-            idx_context = binding.path.get_idx_context()
+            idx_context = binding.path.index_context
             if idx_context in out.keys():
                 continue
             out[idx_context] = tree
@@ -855,7 +1089,7 @@ def tree_for_idxcontext(
     for tree in copy.deepcopy(list(out.values())):
         for level_name in tree.level_names[:-1]:
             new_tree = tree.clip_at_level(level_name, inplace=False)
-            idx_context = new_tree.get_idx_context()
+            idx_context = new_tree.index_context
             if idx_context in out.keys():
                 continue
             out[idx_context] = new_tree
