@@ -1,4 +1,5 @@
 from __future__ import annotations
+from pprint import pprint
 from hubit.utils import is_digit
 import os
 import json
@@ -6,9 +7,11 @@ import hashlib
 import logging
 import multiprocessing
 from multiprocessing.managers import SyncManager
+from multiprocessing import Value
+from ctypes import c_bool
 import copy
 from typing import Any, Callable, Dict, Set, TYPE_CHECKING, List, Optional, Union
-from .config import HubitBinding, HubitQueryPath, ModelIndexSpecifier
+from .config import FlatData, HubitBinding, HubitQueryPath, ModelIndexSpecifier
 from .tree import LengthTree
 from .utils import traverse, reshape, ReadOnlyDict
 from operator import itemgetter
@@ -16,7 +19,6 @@ from operator import itemgetter
 from .errors import HubitError, HubitWorkerError
 
 if TYPE_CHECKING:
-    from .qrun import _QueryRunner
     from .config import HubitModelComponent
 
 
@@ -29,7 +31,8 @@ class _Worker:
 
     def __init__(
         self,
-        qrun: _QueryRunner,
+        callback_ready: Callable,
+        callback_completed: Callable,
         component: HubitModelComponent,
         query: HubitQueryPath,
         func: Callable,
@@ -51,7 +54,8 @@ class _Worker:
         self.id = component.id  # name of the component
         self.name = component.name  # name of the component
         self.version = version  # Version of the component
-        self.qrun = qrun  # reference to the query runner
+        self._callback_ready = callback_ready
+        self._callback_completed = callback_completed
         self.job = None  # For referencing the job if using multiprocessing
         self.query = query
         self.tree_for_idxcontext = tree_for_idxcontext
@@ -61,16 +65,19 @@ class _Worker:
         self._consumed_results_ready = False
         self._consumes_input_only = False
         self._results_id: Optional[str] = None
+        self.input_checksum: Optional[str] = None
+        self.results_checksum: Optional[str] = None
         self.caching = caching
+        self._did_start = False
 
         # Store information on how results were created (calculation or cache)
         self._results_from = self.RESULTS_FROM_UNKNOWN
 
         if dryrun:
-            # If worker should perform a dry run set the worker function to "work_dryrun"
-            self.workfun = self.work_dryrun
+            # If worker should perform a dry run set the worker function to "_work_dryrun"
+            self.workfun = self._work_dryrun
         else:
-            self.workfun = self.work
+            self.workfun = self._work
 
         # Paths for values that are consumed but not ready
         self.pending_input_paths: List[HubitQueryPath] = []
@@ -92,9 +99,11 @@ class _Worker:
         self.results: Dict[str, Any]
         if manager is None:
             self.results = {}
+            self._did_complete = False
             self.use_multiprocessing = False
         else:
             self.results = manager.dict()
+            self._did_complete = Value(c_bool, False)
             self.use_multiprocessing = True
 
         # TODO
@@ -174,7 +183,11 @@ class _Worker:
                 for path in traverse(paths)
             }
 
-        logging.info(f'Worker "{self.id}" was deployed for query "{self.query}"')
+        logging.info(f'Worker "{self.id}" was created for query "{self.query}"')
+
+    @property
+    def status(self):
+        return "Not implemented"
 
     @staticmethod
     def bindings_from_idxs(bindings: List[HubitBinding], idxval_for_idxid) -> Dict:
@@ -247,7 +260,7 @@ class _Worker:
         # matched the query suffice
         idxval_for_idxid = {}
         for binding in bindings:
-            if query_path.check_path_match(binding.path):
+            if query_path.path_match(binding.path):
                 identifiers = binding.path.get_index_identifiers()
                 ranges = query_path.ranges()
                 idxval_for_idxid.update(dict(zip(identifiers, ranges)))
@@ -261,7 +274,7 @@ class _Worker:
     def expand(path_for_name, tree_for_idxcontext, model_path_for_name):
         paths_for_name = {}
         for name, path in path_for_name.items():
-            tree = tree_for_idxcontext[model_path_for_name[name].get_idx_context()]
+            tree = tree_for_idxcontext[model_path_for_name[name].index_context]
             pruned_tree = tree.prune_from_path(path, inplace=False)
             paths_for_name[name] = pruned_tree.expand_path(path)
         return paths_for_name
@@ -337,17 +350,6 @@ class _Worker:
         """
         return set(self.results.keys()) == set(self.rpath_provided_for_name.keys())
 
-    def work_dryrun(self):
-        """
-        Sets all results to None
-        """
-        self.qrun._set_worker_working(self)
-        for name in self.rpath_provided_for_name.keys():
-            tree = self.tree_for_idxcontext[
-                self.provided_mpath_for_name[name].get_idx_context()
-            ]
-            self.results[name] = tree.none_like()
-
     def join(self):
         """Join process"""
         if self.job is not None:
@@ -356,21 +358,51 @@ class _Worker:
 
     def use_cached_result(self, result):
         logging.info(f'Worker "{self.id}" using CACHE for query "{self.query}"')
-        self.qrun._set_worker_working(self)
         # Set each key-val pair from the cached results to the worker results
         # The worker results may be a managed dict
         for key, val in result.items():
             self.results[key] = val
         self._results_from = self.RESULTS_FROM_CACHE_ID
 
-    def work(self):
+    def _func_wrapped(self, inputval_for_name, results, did_complete):
+        self.func(inputval_for_name, results)
+        with did_complete.get_lock():
+            did_complete.value = Value(c_bool, True)
+        self._callback_completed(self)
+
+    def _mp_func(self, inputval_for_name):
+        self.job = multiprocessing.Process(
+            target=self._func_wrapped,
+            args=(inputval_for_name, self.results, self._did_complete),
+        )
+        self.job.daemon = False
+        self.job.start()
+
+    def _sp_func(self, inputval_for_name):
+        self.func(inputval_for_name, self.results)
+        self._did_complete = True
+        self._callback_completed(self)
+
+    def _work_dryrun(self):
+        """
+        Sets all results to None
+        """
+        for name in self.rpath_provided_for_name.keys():
+            tree = self.tree_for_idxcontext[
+                self.provided_mpath_for_name[name].index_context
+            ]
+            self.results[name] = tree.none_like()
+
+        self._did_complete = True
+        self._callback_completed(self)
+
+    def _work(self):
         """
         Executes actual work
         """
         logging.info(f'Worker "{self.id}" STARTED for query "{self.query}"')
 
         # Notify the hubit model that we are about to start the work
-        self.qrun._set_worker_working(self)
         # create single input
         inputval_for_name = ReadOnlyDict(
             {
@@ -379,14 +411,9 @@ class _Worker:
             }
         )
         if self.use_multiprocessing:
-            self.job = multiprocessing.Process(
-                target=self.func,
-                args=(inputval_for_name, self.results),
-            )
-            self.job.daemon = False
-            self.job.start()
+            self._mp_func(inputval_for_name)
         else:
-            self.func(inputval_for_name, self.results)
+            self._sp_func(inputval_for_name)
         self._results_from = self.RESULTS_FROM_CALCULATION_ID
 
         logging.debug("\n**STOP WORKING**\n{}".format(self.__str__()))
@@ -405,22 +432,24 @@ class _Worker:
     def is_ready_to_work(self):
         return self._consumed_input_ready and self._consumed_results_ready
 
-    def work_if_ready(self, results=None):
-        """
-        If all consumed attributes are present start working
-        """
-        if self.is_ready_to_work():
-            logging.debug("Let the work begin: {}".format(self.workfun))
+    def _prepare_work(self):
+        logging.debug("Let the work begin: {}".format(self.workfun))
+        self._consumed_data_set = True
+        self._did_start = True
 
-            self.resultval_for_name = _Worker.reshape(
-                self.rpaths_consumed_for_name, self.resultval_for_path
-            )
+    def work(self):
+        """ """
+        self._prepare_work()
+        self.workfun()
 
-            self._consumed_data_set = True
-            if results is None:
-                self.workfun()
-            else:
-                self.use_cached_result(results)
+    def set_results(self, results):
+        """Set pre-computed results directly on worker.
+        Don't wait for input.
+        """
+        self._prepare_work()
+        self.use_cached_result(results)
+        self._did_complete = True
+        self._callback_completed(self)
 
     def set_consumed_input(self, path: HubitQueryPath, value):
         if path in self.pending_input_paths:
@@ -434,6 +463,16 @@ class _Worker:
                 self.ipaths_consumed_for_name, self.inputval_for_path
             )
 
+        if len(self.pending_input_paths) == 0:
+            self.input_checksum = self._input_checksum()
+
+        if self.is_ready_to_work():
+            self.resultval_for_name = _Worker.reshape(
+                self.rpaths_consumed_for_name, self.resultval_for_path
+            )
+            self.results_checksum = self._results_checksum()
+            self._callback_ready(self)
+
         # if not self.caching:
         #     self.work_if_ready()
 
@@ -443,17 +482,20 @@ class _Worker:
             self.resultval_for_path[path] = value
 
         self._consumed_results_ready = len(self.pending_results_paths) == 0
-        self.work_if_ready()
 
-    def set_values(self, inputdata, resultsdata):
+        if self.is_ready_to_work():
+            self.resultval_for_name = _Worker.reshape(
+                self.rpaths_consumed_for_name, self.resultval_for_path
+            )
+            self.results_checksum = self._results_checksum()
+
+            self._callback_ready(self)
+
+    def set_values(self, inputdata, resultsdata: FlatData):
         """
         Set the consumed values if they are ready otherwise add them
         to the list of pending items
         """
-        # set the worker here since in init we have not yet
-        # checked that a similar instance does not exist
-        self.qrun._set_worker(self)
-
         # Check consumed input (should not have any pending items by definition)
         for path in self.iname_for_path.keys():
             if path in inputdata.keys():
@@ -476,6 +518,16 @@ class _Worker:
             self.inputval_for_name = _Worker.reshape(
                 self.ipaths_consumed_for_name, self.inputval_for_path
             )
+            self.input_checksum = self._input_checksum()
+
+        if self._consumed_results_ready:
+            self.resultval_for_name = _Worker.reshape(
+                self.rpaths_consumed_for_name, self.resultval_for_path
+            )
+            self.results_checksum = self._results_checksum()
+
+        if self.is_ready_to_work():
+            self._callback_ready(self)
 
         return (
             copy.copy(self.pending_input_paths),
@@ -493,8 +545,8 @@ class _Worker:
             "&".join([f"{k}={v}" for k, v in self.idxval_for_idxid.items()]),
         )
 
-    def _make_results_id(self):
-        """results_ids based on input and function only"""
+    def _input_checksum(self) -> str:
+        """checksum for the input values"""
         return hashlib.md5(
             # f'{self.inputval_for_name}_{id(self.func)}'.encode('utf-8')
             json.dumps(
@@ -506,42 +558,29 @@ class _Worker:
             ).encode()
         ).hexdigest()
 
-    def set_results_id(self, results_ids: List[str]) -> str:
-        """results_ids are the IDs of workers spawned from the
-        current worker
-
-        augment that with worker's own results_id
+    def _results_checksum(self) -> str:
         """
-        results_ids.append(self._make_results_id())
-        self._results_id = hashlib.md5("".join(results_ids).encode("utf-8")).hexdigest()
-        return self._results_id
-
-    @property
-    def results_id(self):
-        return self._results_id or self._set_results_id()
-
-    def _set_results_id(self):
-        """checksum for worker function and input. This ID is identical for
-        worker that carry out the same calculation
+        We want to know the checksum before the results values are available.
+        The results will be identifiable by
+          - the worker input (input_checksum)
+          - the consumed results
         """
-        if self._results_id is not None:
-            return self._results_id
-
-        if not self._consumes_input_only:
-            raise HubitWorkerError("Not safe to create results ID")
-
-        if not self._consumed_input_ready:
-            raise HubitWorkerError("Not enough data to create calc ID")
-
-        self._results_id = self._make_results_id()
-        return self._results_id
+        return hashlib.md5(
+            json.dumps(
+                [
+                    self.input_checksum,
+                    sorted(self.resultval_for_name.items(), key=itemgetter(0)),
+                ]
+            ).encode()
+        ).hexdigest()
 
     def __str__(self):
         n = 100
         fstr1 = "{:30}{}\n"
         strtmp = "=" * n + "\n"
-        strtmp += "ID {}\n".format(self.idstr())
-        strtmp += "Function {}\n".format(self.func)
+        strtmp += "ID: {}\n".format(self.idstr())
+        strtmp += "Function: {}\n".format(self.func)
+        strtmp += "Query: {}\n".format(self.query)
         strtmp += "-" * n + "\n"
         strtmp += fstr1.format("Results provided", self.rpath_provided_for_name)
         strtmp += fstr1.format(
@@ -563,7 +602,11 @@ class _Worker:
         strtmp += fstr1.format("Results pending", self.pending_results_paths)
 
         strtmp += "-" * n + "\n"
+        strtmp += f"Ready to work: {self.is_ready_to_work()}\n"
+        strtmp += f"Did start: {self._did_start}\n"
+        strtmp += f"Did complete: {self._did_complete}\n"
         strtmp += "Results {}\n".format(self.results)
+        strtmp += f"Results from: {self._results_from}\n"
 
         strtmp += "=" * n + "\n"
 
