@@ -196,6 +196,73 @@ class TestHubitComponent(unittest.TestCase):
         cfg.update({"index_scope": {"IDX": ":"}})
         HubitModelComponent.from_cfg(cfg, 0)
 
+    def test_get_paths(self):
+        cfg = {
+            "path": "dummy",
+            "func_name": "dummy",
+            "provides_results": [
+                {"name": "attr", "path": "list[IDX].attr.path2"},
+            ],
+            "consumes_input": [
+                {"name": "attr1", "path": "shared.input.attr.path1"},
+                {"name": "attr2", "path": "shared.input.attr.path2"},
+            ],
+            "consumes_results": [
+                {"name": "attr3", "path": "shared.input.attr.path3"},
+                {"name": "attr4", "path": "shared.input.attr.path4"},
+            ],
+        }
+
+        cmp = HubitModelComponent.from_cfg(cfg, 0)
+        result = set(cmp.consumes_input_paths)
+        expected_result = set(
+            [
+                HubitModelPath("shared.input.attr.path1"),
+                HubitModelPath("shared.input.attr.path2"),
+            ]
+        )
+        assert result == expected_result
+
+        result = set(cmp.consumes_results_paths)
+        expected_result = set(
+            [
+                HubitModelPath("shared.input.attr.path3"),
+                HubitModelPath("shared.input.attr.path4"),
+            ]
+        )
+        assert result == expected_result
+
+        result = set(cmp.provides_results_paths)
+        expected_result = set(
+            [
+                HubitModelPath("list[IDX].attr.path2"),
+            ]
+        )
+        assert result == expected_result
+
+    def test_scope_start(self):
+        cfg = {
+            "path": "dummy",
+            "func_name": "dummy",
+            "provides_results": [
+                {"name": "attr", "path": "list[IDX].attr.path2"},
+            ],
+        }
+
+        # No scope
+        cmp = HubitModelComponent.from_cfg(cfg, 0)
+        assert cmp.scope_start == (None, None)
+
+        # All slice
+        cfg.update({"index_scope": {"IDX": "1:4"}})
+        cmp = HubitModelComponent.from_cfg(cfg, 0)
+        assert cmp.scope_start == ("IDX", 1)
+
+        # All indices
+        cfg.update({"index_scope": {"IDX": ":"}})
+        cmp = HubitModelComponent.from_cfg(cfg, 0)
+        assert cmp.scope_start == ("IDX", 0)
+
 
 class TestHubitPath(unittest.TestCase):
     def test_from_dotted(self):
@@ -273,6 +340,61 @@ class TestHubitQueryPath(unittest.TestCase):
         idxs_match = qpath.idxs_for_matches(mpaths)
         self.assertSequenceEqual(idxs_match, idxs_match_expected)
 
+    def test_path_match(self):
+        # Match
+        qpath = HubitQueryPath("segs[42].walls[3].temps")
+        mpath = HubitModelPath("segs[IDX1].walls[IDX1].temps")
+        assert qpath.path_match(mpath)
+
+        # Match
+        mpath = HubitModelPath("segs[:@IDX1].walls[IDX1].temps")
+        assert qpath.path_match(mpath)
+
+        # No match: different field names
+        mpath = HubitModelPath("seg[:@IDX1].wall[IDX1].temp")
+        assert not qpath.path_match(mpath)
+
+        # No match: different field count
+        qpath = HubitQueryPath("segs[42].walls[3].temps")
+        mpath = HubitModelPath("segs[IDX1].walls[IDX1]")
+        assert not qpath.path_match(mpath)
+
+        # No match: different brace count
+        qpath = HubitQueryPath("segs[42].walls[3].temps")
+        mpath = HubitModelPath("segs.walls[IDX1].temps")
+        assert not qpath.path_match(mpath)
+
+        # No match: no intersection for second index
+        qpath = HubitQueryPath("segs[42].walls[3].temps")
+        mpath = HubitModelPath("segs[IDX1].walls[43@IDX2].temps")
+        assert not qpath.path_match(mpath)
+
+        ### Negative index in query
+        # Match: negative index in query
+        qpath = HubitQueryPath("segs[42].walls[-1].temps")
+        mpath = HubitModelPath("segs[IDX1].walls[IDX1].temps")
+        assert qpath.path_match(mpath)
+
+        # Match: negative index in query
+        mpath = HubitModelPath("segs[IDX1].walls[:@IDX1].temps")
+        assert qpath.path_match(mpath)
+
+        # No match: negative index in query. Index 12 might be the last index, i.e. math -1 but we cannot guarantee it
+        mpath = HubitModelPath("segs[IDX1].walls[12@IDX1].temps")
+        assert not qpath.path_match(mpath)
+
+    def test_new_with_index(self):
+        # Replace only first occurrence
+        path = HubitModelPath("segments[1].lay1ers[1]")
+        path = path.new_with_index("1", "11")
+        assert path == HubitModelPath("segments[11].lay1ers[1]")
+
+        # Do not change a match in the field name (1 in lay1ers)
+        path = HubitModelPath("segments[2].lay1ers[1]")
+        path = path.new_with_index("1", "11")
+        print(path)
+        assert path == HubitModelPath("segments[2].lay1ers[11]")
+
 
 class TestHubitModelPath(unittest.TestCase):
     def test_has_slice_range(self):
@@ -306,7 +428,7 @@ class TestHubitModelPath(unittest.TestCase):
 
     def test_get_idx_context(self):
         path = HubitModelPath("segs[:@IDX_SEG].walls[IDX_WALL].heat_flow")
-        result = path.get_idx_context()
+        result = path.index_context
         expected_result = "IDX_SEG-IDX_WALL"
         self.assertSequenceEqual(result, expected_result)
 
@@ -389,14 +511,13 @@ class TestHubitModelPath(unittest.TestCase):
         path = HubitModelPath("segments[17@IDX_SEG].layers[0@IDX_LAY]")
         path.validate()
 
-        # Valid (limited index range)
+        # Invalid since limited index ranges are not allowed
         path = HubitModelPath("segments[17:34@IDX_SEG].layers[IDX_LAY]")
         path.validate()
 
-        # Invalid due to negative index range
+        # Valid. Negative index range allowed
         path = HubitModelPath("segments[-1@IDX_SEG].layers[IDX_LAY]")
-        with self.assertRaises(HubitError):
-            path.validate()
+        path.validate()
 
         # Only one @ allowed in index specifier
         path = HubitModelPath("segments[IDX_SEG].layers[:@@IDX_LAY]")
@@ -430,6 +551,12 @@ class TestHubitModelPath(unittest.TestCase):
         path = HubitModelPath("segments[IDX_SEG].layers[:@IDX1LAY113]")
         assert path.as_query_depth_path() == "segments[*].layers[*]"
 
+    def test_new_with_index(self):
+        path = HubitModelPath("segments[IDX_SEG].lay1ers[:@IDX1LAY113]")
+        index_specifiers = path.get_index_specifiers()
+        path = path.new_with_index(index_specifiers[1], "1")
+        assert path == HubitModelPath("segments[IDX_SEG].lay1ers[1]")
+
 
 class TestModelIndexSpecifier(unittest.TestCase):
     def test_mis(self):
@@ -633,7 +760,7 @@ class TestPathIndexRange(unittest.TestCase):
 
     def test_invalid(self):
         with self.assertRaises(HubitError):
-            PathIndexRange("-1")
+            PathIndexRange("-1", allow_negative_index=False)
 
         with self.assertRaises(HubitError):
             PathIndexRange("k")
