@@ -4,7 +4,7 @@ import os
 import pathlib
 import yaml
 from hubit.errors import HubitModelNoInputError, HubitModelQueryError
-from hubit.config import HubitModelConfig, HubitModelPath
+from hubit.config import HubitModelConfig, HubitModelPath, HubitQueryPath
 from hubit import HubitModel
 import pprint
 
@@ -18,6 +18,60 @@ TMP_DIR = os.path.join(THIS_DIR, REL_TMP_DIR)
 pathlib.Path(TMP_DIR).mkdir(parents=True, exist_ok=True)
 
 
+def get_model_with_explicit_indices():
+    return """
+    components:
+        -   
+          # move a number
+          func_name: move_number
+          path: ./components/comp0.py 
+          provides_results:
+            - name: number
+              path: first_coor[IDX1].second_coor[0@IDX2].value 
+          consumes_input: 
+            - name: number
+              path: list[IDX1].some_attr.inner_list[IDX2].xval
+        -
+          func_name: move_number
+          path: ./components/comp0.py 
+          provides_results:
+            - name: number
+              path: first_coor[IDX1].second_coor[1@IDX2].value 
+          consumes_input: 
+            - name: number
+              path: list[IDX1].some_attr.inner_list[IDX2].xval
+    """
+
+
+def get_model_with_index_scopes():
+    return """
+    components:
+        -
+          # move a number
+          func_name: move_number
+          path: ./components/comp0.py
+          index_scope:
+            IDX2: 0
+          provides_results:
+            - name: number
+              path: first_coor[IDX1].second_coor[IDX2].value
+          consumes_input:
+            - name: number
+              path: list[IDX1].some_attr.inner_list[IDX2].xval
+        -
+          func_name: move_number
+          path: ./components/comp0.py
+          index_scope:
+            IDX2: "1:"
+          provides_results:
+            - name: number
+              path: first_coor[IDX1].second_coor[IDX2].value
+          consumes_input:
+            - name: number
+              path: list[IDX1].some_attr.inner_list[IDX2].xval
+    """
+
+
 def setUpModule():
     global yml_input
     global model
@@ -68,6 +122,109 @@ class TestModel(unittest.TestCase):
 
         self.querystr_level0_last = "list[-1].some_attr.two_x_numbers"
 
+    def run_mpaths_for_qpath_fields_only(hmodel: HubitModel):
+
+        expected_mpaths = [
+            "first_coor[IDX1].second_coor[0@IDX2].value",
+            "first_coor[IDX1].second_coor[1@IDX2].value",
+        ]
+
+        expected_cmp_ids = [
+            "cmp0@./components/comp0.move_number",
+            "cmp1@./components/comp0.move_number",
+        ]
+
+        # Two components match query path
+        qpath = HubitQueryPath("first_coor[:].second_coor[:].value")
+        mpaths, cmp_ids = hmodel.mpaths_for_qpath_fields_only(qpath)
+
+        assert cmp_ids == expected_cmp_ids
+        assert mpaths == expected_mpaths
+
+        # Only the second components match query path
+        qpath = HubitQueryPath("first_coor[:].second_coor[1].value")
+        mpaths, cmp_ids = hmodel.mpaths_for_qpath_fields_only(qpath)
+
+        assert cmp_ids == expected_cmp_ids
+        assert mpaths == expected_mpaths
+
+        # Both components match the query path since we don't check the intersection
+        qpath = HubitQueryPath("first_coor[:].second_coor[1].value")
+        mpaths, cmp_ids = hmodel.mpaths_for_qpath_fields_only(qpath)
+
+        assert cmp_ids == expected_cmp_ids
+        assert mpaths == expected_mpaths
+
+    def test_mpaths_for_qpath_fields_only(self):
+        # Test the default model
+        qpath = HubitQueryPath("first_coor[:].second_coor[:].value")
+        mpaths, cmp_ids = self.hmodel.mpaths_for_qpath_fields_only(qpath)
+        expected_cmp_ids = [
+            "cmp0@./components/comp0.move_number",
+        ]
+        assert cmp_ids == expected_cmp_ids
+
+        expected_mpaths = ["first_coor[IDX1].second_coor[IDX2].value"]
+        assert mpaths == expected_mpaths
+
+        model = get_model_with_explicit_indices()
+        model_cfg = HubitModelConfig.from_cfg(
+            yaml.load(model, Loader=yaml.FullLoader), base_path=THIS_DIR
+        )
+        hmodel = HubitModel(model_cfg, name="TEST", output_path=REL_TMP_DIR)
+        TestModel.run_mpaths_for_qpath_fields_only(hmodel)
+
+    def run_cmpids_for_query_tests(hmodel: HubitModel):
+        # Two components match query path
+        qpath = HubitQueryPath("first_coor[:].second_coor[:].value")
+        result = hmodel._cmpids_for_query(qpath, check_intersection=True)
+        expected_result = [
+            "cmp0@./components/comp0.move_number",
+            "cmp1@./components/comp0.move_number",
+        ]
+        assert result == expected_result
+
+        # Only the second components match query path
+        qpath = HubitQueryPath("first_coor[:].second_coor[1].value")
+        result = hmodel._cmpids_for_query(qpath, check_intersection=True)
+        expected_result = [
+            "cmp1@./components/comp0.move_number",
+        ]
+        assert result == expected_result
+
+        # Both components match the query path since we don't check the intersection
+        qpath = HubitQueryPath("first_coor[:].second_coor[1].value")
+        result = hmodel._cmpids_for_query(qpath, check_intersection=False)
+        expected_result = [
+            "cmp0@./components/comp0.move_number",
+            "cmp1@./components/comp0.move_number",
+        ]
+        assert result == expected_result
+
+    def test_cmpids_for_query(self):
+
+        # Test the default model
+        qpath = HubitQueryPath("first_coor[:].second_coor[:].value")
+        result = self.hmodel._cmpids_for_query(qpath, check_intersection=True)
+        expected_result = ["cmp0@./components/comp0.move_number"]
+        assert result == expected_result
+
+        # Test model with explicit indices
+        model = get_model_with_explicit_indices()
+        model_cfg = HubitModelConfig.from_cfg(
+            yaml.load(model, Loader=yaml.FullLoader), base_path=THIS_DIR
+        )
+        hmodel = HubitModel(model_cfg, name="TEST", output_path=REL_TMP_DIR)
+        TestModel.run_cmpids_for_query_tests(hmodel)
+
+        # Test model with index scopes
+        model = get_model_with_index_scopes()
+        model_cfg = HubitModelConfig.from_cfg(
+            yaml.load(model, Loader=yaml.FullLoader), base_path=THIS_DIR
+        )
+        hmodel = HubitModel(model_cfg, name="TEST", output_path=REL_TMP_DIR)
+        TestModel.run_cmpids_for_query_tests(hmodel)
+
     def test_from_file(self):
         """
         Test if model can successfully be loaded from a file
@@ -136,9 +293,6 @@ class TestModel(unittest.TestCase):
         """
         Render the query
         TODO: could test the dot object instead
-        Sometimes (?!?!) gives this warning:
-        Warning: _Query -> cluster_resultslist: head not inside head cluster cluster_results
-        Warning: cluster_resultslist -> _Response: tail not inside tail cluster cluster_results
         """
         self.hmodel.set_input(self.input)
         query = [self.querystr_level0]
@@ -282,33 +436,6 @@ class TestModel(unittest.TestCase):
             with self.subTest(use_multi_processing=use_multi_processing):
                 test()
 
-    def test_get_last(self):
-        """
-        Query last list element
-        """
-        self.skipTest("Catch22 in normalize, prune, expand")
-        self.hmodel.set_input(self.input)
-        queries = [self.querystr_level0_last]
-        for use_multi_processing in self.use_multi_processing_values:
-            with self.subTest(use_multi_processing=use_multi_processing):
-                response = self.hmodel.get(
-                    queries, use_multi_processing=use_multi_processing
-                )
-                self.assertSequenceEqual(
-                    response[self.querystr_level0_last], self.expected_result_level0
-                )
-
-    def test_get_all(self):
-        """
-        No query yields all results
-        """
-        self.skipTest("Not implemented")
-        self.hmodel.set_input(self.input)
-        for use_multi_processing in self.use_multi_processing_values:
-            with self.subTest(use_multi_processing=use_multi_processing):
-                response = self.hmodel.get(use_multi_processing=use_multi_processing)
-                print(response)
-
     def test_sweep(self):
         """
         Sweep input parameters
