@@ -1,17 +1,22 @@
 import os
 import unittest
+import pytest
 import yaml
 
 from hubit.model import HubitModel
-from hubit.qrun import _QueryRunner
+from hubit.qrun import _QueryRunner, query_runner_factory
 from hubit.errors import HubitModelQueryError
-from hubit.config import FlatData, HubitModelConfig, HubitQueryPath
+from hubit.config import FlatData, HubitModelComponent, HubitModelConfig, HubitQueryPath
 
 THIS_FILE = os.path.realpath(__file__)
 THIS_DIR = os.path.dirname(THIS_FILE)
 REL_TMP_DIR = "./tmp"
 
 
+def function_for_test():
+    ...
+
+
 def setUpModule():
     global yml_input
     global model
@@ -72,7 +77,7 @@ class TestRunner(unittest.TestCase):
             output_path=REL_TMP_DIR,
         )
         use_multi_processing = False
-        self.qr = _QueryRunner(self.hmodel, use_multi_processing)
+        self.qr = query_runner_factory(use_multi_processing, self.hmodel)
         self.input = yaml.load(yml_input, Loader=yaml.FullLoader)
         self.hmodel.set_input(self.input)
 
@@ -85,6 +90,83 @@ class TestRunner(unittest.TestCase):
             "list[{}].some_attr.two_x_numbers_x_factor".format(self.idx)
         )
 
+    def test_get_func(self):
+
+        # Location of model file
+        base_path = THIS_DIR
+
+        cfg = {
+            "path": "dummy",
+            "func_name": "dummy_func",
+            "provides_results": [
+                {"name": "attr", "path": "list[IDX].attr.path1"},
+            ],
+        }
+
+        component = HubitModelComponent.from_cfg(cfg, 0)
+
+        # Components already known
+        components_known = {}
+
+        # cannot find module "dummy"
+        with pytest.raises(AssertionError):
+            _QueryRunner._get_func(base_path, component, components_known)
+
+        cfg["path"] = "config_test.py"
+        component = HubitModelComponent.from_cfg(cfg, 0)
+
+        # Cannot find attribute "dummy_func"
+        with pytest.raises(AttributeError):
+            _QueryRunner._get_func(base_path, component, components_known)
+
+        cfg["path"] = "qrun_test.py"
+        cfg["func_name"] = "function_for_test"
+        component = HubitModelComponent.from_cfg(cfg, 0)
+        func, version, components_known = _QueryRunner._get_func(
+            base_path, component, components_known
+        )
+        assert func.__name__ == function_for_test.__name__
+        assert func.__module__ == function_for_test.__module__.split(".")[-1]
+        assert version == "None"
+        assert components_known == {component.id: (func, version)}
+
+        # Run a second time so that we test components_known is not empty
+        func, version, components_known = _QueryRunner._get_func(
+            base_path, component, components_known
+        )
+        assert func.__name__ == function_for_test.__name__
+        assert func.__module__ == function_for_test.__module__.split(".")[-1]
+        assert version == "None"
+        assert components_known == {component.id: (func, version)}
+
+        cfg["path"] = "math"
+        cfg["func_name"] = "cos"
+        cfg["is_dotted_path"] = True
+        components_known = {}
+        component = HubitModelComponent.from_cfg(cfg, 0)
+        func, version, components_known = _QueryRunner._get_func(
+            base_path, component, components_known
+        )
+        print(func.__module__)
+        assert func.__name__ == "cos"
+        assert func.__module__ == "math"
+        assert version == "None"
+        assert components_known == {component.id: (func, version)}
+
+    def test_str(self):
+        """Check the string representation completes"""
+        # Use query to set worker(s) to cover more code blocks in __str__
+        self.get_worker_counts([[self.querystr_level0]])
+        str(self.qr)
+
+    def test_status(self):
+        """Check the status string representation completes"""
+        self.get_worker_counts([[self.querystr_level0]])
+        flat_results = FlatData({"hello.world": 5})
+        self.qr.flat_results = flat_results
+        expanded_paths = [HubitQueryPath("hello.world")]
+        self.qr._status(flat_results, expanded_paths)
+
     def test_worker_comp1(self):
         """ """
         # Component index in model (TODO: brittle)'
@@ -133,17 +215,15 @@ class TestRunner(unittest.TestCase):
             self.qr._worker_for_query(HubitQueryPath("i.dont.exist"))
 
     def get_worker_counts(self, queries):
-        flat_results = FlatData()
         flat_input = FlatData.from_dict(
             self.input,
             stop_at=self.model_cfg.compiled_query_depths,
             include_patterns=self.model_cfg.include_patterns,
         )
+        self.qr.reset()
         worker_counts = []
         for qpaths in queries:
-            self.qr.spawn_workers(
-                qpaths, flat_input, flat_results, flat_input, dryrun=True
-            )
+            self.qr.spawn_workers(qpaths, flat_input, flat_input, dryrun=True)
             worker_counts.append(len(self.qr.workers))
 
         return worker_counts
