@@ -1,4 +1,7 @@
+import random
+import string
 import unittest
+from unittest.mock import Mock
 import yaml
 import pytest
 import pprint
@@ -10,8 +13,48 @@ from hubit.tree import (
     _QueryExpansion,
     LeafNode,
 )
-from hubit.config import HubitModelPath, HubitQueryPath
-from hubit.errors import HubitIndexError, HubitModelQueryError
+from hubit.config import FlatData, HubitModelPath, HubitQueryPath, PathIndexRange
+from hubit.errors import HubitError, HubitIndexError, HubitModelQueryError
+
+
+def get_mock_components(mpaths):
+    cmp = Mock()
+    cmp.index_scope = {}
+    return [cmp for _ in mpaths]
+
+
+def _get_data():
+    # Make the tree
+    site_nodes = LengthNode(2)
+    line_nodes = [LengthNode(1), LengthNode(1)]
+    tank_nodes_site0 = [LengthNode(3)]
+    tank_nodes_site1 = [LengthNode(4)]
+
+    site_nodes.set_children(line_nodes)
+    line_nodes[0].set_children(tank_nodes_site0)
+    line_nodes[1].set_children(tank_nodes_site1)
+
+    nodes = [site_nodes]
+    nodes.extend(line_nodes)
+    nodes.extend(tank_nodes_site0)
+    nodes.extend(tank_nodes_site1)
+    level_names = "IDX_SITE", "IDX_LINE", "IDX_TANK"
+
+    flat_results = FlatData(
+        {
+            "sites[0].lines[0].tanks[0].Q_yield": 10.0,
+            "sites[0].lines[0].tanks[1].Q_yield": 6.0,
+            "sites[0].lines[0].tanks[2].Q_yield": 4.0,
+            "sites[1].lines[0].tanks[0].Q_yield": 20.0,
+            "sites[1].lines[0].tanks[1].Q_yield": 8.0,
+            "sites[1].lines[0].tanks[2].Q_yield": 3.0,
+            "sites[1].lines[0].tanks[3].Q_yield": 1.0,
+        }
+    )
+    tree = LengthTree(nodes, level_names)
+    print("tree", tree)
+
+    return tree, flat_results
 
 
 class TestDummyTree(unittest.TestCase):
@@ -23,12 +66,12 @@ class TestDummyTree(unittest.TestCase):
 
     def test_get_idx_context(self):
         """Index context for dummy tree must be the empty string"""
-        result = self.tree.get_idx_context()
+        result = self.tree.index_context
         assert result == ""
 
     def test_prune_from_path(self):
         """Inplace pruning gives the dummy tree itself"""
-        result = self.tree.prune_from_path(inplace=True)
+        result = self.tree.prune_from_path(HubitModelPath(), inplace=True)
         assert result == self.tree
 
     def test_clip_at_level(self):
@@ -41,6 +84,14 @@ class TestDummyTree(unittest.TestCase):
         result = self.tree.fix_idx_at_level()
         assert result == self.tree
 
+    def test_is_path_described(self):
+        """Any path is describe"""
+        assert DummyLengthTree().is_path_described(
+            HubitModelPath(
+                "".join(random.choices(string.ascii_uppercase + string.digits, k=10))
+            )
+        )
+
 
 class TestTree(unittest.TestCase):
     def setUp(self):
@@ -347,39 +398,6 @@ class TestTree(unittest.TestCase):
         ]
         self.assertListEqual(tree.to_list(), expected_lengths)
 
-    def test_normalize_path1(self):
-        """
-        First index ID is negative so no context
-        """
-        qpath = HubitModelPath("segments[-1].layers[XX].test.positions[YY]")
-        normalized_qpath_expected = "segments[1].layers[XX].test.positions[YY]"
-        normalized_qpath = self.tree.normalize_path(qpath)
-        self.assertEqual(normalized_qpath_expected, normalized_qpath)
-
-    def test_normalize_path2(self):
-        """
-        Second index ID is negative there's context
-        """
-        qpaths = (
-            HubitModelPath("segments[0].layers[-1].test.positions[:]"),
-            HubitModelPath("segments[1].layers[-1].test.positions[:]"),
-        )
-        normalized_qpaths_expected = (
-            "segments[0].layers[2].test.positions[:]",
-            "segments[1].layers[3].test.positions[:]",
-        )
-
-        for qpath, expected_qpath in zip(qpaths, normalized_qpaths_expected):
-            path = self.tree.normalize_path(qpath)
-            with self.subTest(path=path, expected_qpath=expected_qpath):
-                self.assertEqual(expected_qpath, path)
-
-    def test_normalize_path3(self):
-        qpath = HubitModelPath("segments[1].layers[3].test.positions[-1]")
-        expected_qpath = "segments[1].layers[3].test.positions[3]"
-        normalized_qpath = self.tree.normalize_path(qpath)
-        self.assertEqual(expected_qpath, normalized_qpath)
-
     def test_expand_mpath1(self):
         """Expand to full template path"""
         path = HubitModelPath(
@@ -427,9 +445,13 @@ class TestTree(unittest.TestCase):
             ],
         ]
 
-        paths = self.tree.expand_path(path)
+        paths = self.tree.prune_from_path(path, inplace=False).expand_path(path)
         self.assertSequenceEqual(paths, expected_paths)
 
+        # Cannot expand without pruning
+        with self.assertRaises(HubitError) as context:
+            paths = self.tree.expand_path(path)
+
     def test_expand_path2(self):
         """Expand path"""
         paths = (
@@ -448,7 +470,10 @@ class TestTree(unittest.TestCase):
         tree = LengthTree(nodes, level_names)
         # 2  values for segment 0 and 2 values for segment 1
         expected_paths = [
-            ["segments[0].layers[0].test", "segments[0].layers[1].test"],
+            [
+                "segments[0].layers[0].test",
+                "segments[0].layers[1].test",
+            ],
             [
                 "segments[1].layers[0].test",
                 "segments[1].layers[1].test",
@@ -457,9 +482,68 @@ class TestTree(unittest.TestCase):
 
         for path in paths:
             with self.subTest(path=path):
-                expanded_paths = tree.expand_path(path)
+                expanded_paths = tree.prune_from_path(path).expand_path(path)
                 self.assertSequenceEqual(expanded_paths, expected_paths)
 
+    def test_expand_path_count_from_back(self):
+        """Expand path with fixed index set to negative number"""
+        path = HubitQueryPath("segments[:].layers[-2].test")
+
+        seg_node = LengthNode(3)
+        lay_nodes = LengthNode(2), LengthNode(4), LengthNode(3)
+        seg_node.set_children(lay_nodes)
+
+        nodes = [seg_node]
+        nodes.extend(lay_nodes)
+        level_names = "IDX_SEG", "IDX_LAY"
+        tree = LengthTree(nodes, level_names)
+
+        expected_paths = [
+            "segments[0].layers[0].test",
+            "segments[1].layers[2].test",
+            "segments[2].layers[1].test",
+        ]
+
+        # Since the tree is not pruned we must use flat=True
+        expanded_paths = tree.prune_from_path(path, inplace=False).expand_path(
+            path, flat=True
+        )
+        self.assertSequenceEqual(expanded_paths, expected_paths)
+
+        # Index error. There are 2 layers on segment at index 0.
+        path = HubitQueryPath("segments[:].layers[-3].test")
+        with self.assertRaises(HubitIndexError) as context:
+            expanded_paths = tree.prune_from_path(path, inplace=False).expand_path(
+                path, flat=True
+            )
+
+        tree, _ = _get_data()
+        qpath = HubitQueryPath("sites[1].lines[0].tanks[-1].Q_yield")
+        expanded_paths = tree.prune_from_path(qpath, inplace=False).expand_path(
+            qpath, flat=True
+        )
+        assert expanded_paths == ["sites[1].lines[0].tanks[3].Q_yield"]
+
+        qpath = HubitQueryPath("sites[0].lines[-1].tanks[-1].Q_yield")
+        expanded_paths = tree.prune_from_path(qpath, inplace=False).expand_path(
+            qpath, flat=True
+        )
+        assert expanded_paths == ["sites[0].lines[0].tanks[2].Q_yield"]
+
+        qpath = HubitQueryPath("sites[:].lines[-1].tanks[-1].Q_yield")
+        expanded_paths = tree.prune_from_path(qpath, inplace=False).expand_path(
+            qpath, flat=True
+        )
+        assert expanded_paths == [
+            "sites[0].lines[0].tanks[2].Q_yield",
+            "sites[1].lines[0].tanks[3].Q_yield",
+        ]
+
+        qpath = HubitQueryPath("sites[:].lines[:].tanks[:].Q_yield")
+        expanded_paths = tree.prune_from_path(qpath, inplace=False).expand_path(
+            qpath, flat=True
+        )
+
     def test_expand_mpath3(self):
         """Prune tree before expanding. Two indices vary so
         expanded paths is 2D
@@ -487,11 +571,6 @@ class TestTree(unittest.TestCase):
         paths = pruned_tree.expand_path(mpath)
         self.assertSequenceEqual(paths, expected_paths)
 
-        qpath = HubitQueryPath("segments[0].layers[:].test.positions[:]")
-        pruned_tree = self.tree.prune_from_path(qpath, inplace=False)
-        paths = pruned_tree.expand_path(mpath)
-        self.assertSequenceEqual(paths, expected_paths)
-
     def test_expand_mpath4(self):
         """Prune tree before expanding. One index varies so
         expanded paths is 1D
@@ -539,7 +618,9 @@ class TestTree(unittest.TestCase):
         }
 
         result = {
-            name: tree.expand_path(path_consumed_for_name[name])
+            name: tree.prune_from_path(path_consumed_for_name[name]).expand_path(
+                path_consumed_for_name[name]
+            )
             for name, tree in tree_for_name.items()
         }
 
@@ -816,6 +897,67 @@ class TestTree(unittest.TestCase):
                     else:
                         self.assertListEqual(result, expected_result)
 
+    def test_is_path_described_query_paths(self):
+        print(self.tree)
+
+        # Path with no ranges is not described by a LengthTree
+        assert not self.tree.is_path_described(HubitQueryPath("i.dont.exist"))
+
+        assert self.tree.is_path_described(
+            HubitQueryPath("segments[0].layers[0].test.positions[0]")
+        )
+
+        assert not self.tree.is_path_described(
+            HubitQueryPath("segments[0].layers[0].test.positions[1]")
+        )
+
+        assert not self.tree.is_path_described(
+            HubitQueryPath("segments[7].layers[0].test.positions[1]")
+        )
+
+        assert self.tree.is_path_described(
+            HubitQueryPath("segments[1].layers[0].test.positions[4]")
+        )
+
+        assert not self.tree.is_path_described(
+            HubitQueryPath("segments[1].layers[0].test.positions[5]")
+        )
+
+        # At least one segment has 4 positions on layer 0
+        assert self.tree.is_path_described(
+            HubitQueryPath("segments[:].layers[0].test.positions[4]")
+        )
+
+        # No segment has 5 positions on layer 0
+        assert not self.tree.is_path_described(
+            HubitQueryPath("segments[:].layers[0].test.positions[5]")
+        )
+
+    def test_is_path_described_model_paths(self):
+        print(self.tree)
+
+        # Path with no ranges is not described by a LengthTree
+        assert not self.tree.is_path_described(HubitModelPath("i.dont.exist"))
+
+        assert self.tree.is_path_described(
+            HubitModelPath(
+                "segments[0@IDX_SEG].layers[0@IDX_LAY].test.positions[0@IDX_POS]"
+            )
+        )
+
+        assert not self.tree.is_path_described(
+            HubitModelPath(
+                "segments[0@IDX_SEG].layers[0@IDX_LAY].test.positions[1@IDX_POS]"
+            )
+        )
+
+        # Wrong index context
+        assert not self.tree.is_path_described(
+            HubitModelPath(
+                "segments[0@IDX_WRONG].layers[0@IDX_LAY].test.positions[0@IDX_POS]"
+            )
+        )
+
 
 # TODO: test if len(mpaths) > 1 and not path.has_slice_range():
 class TestQueryExpansion(unittest.TestCase):
@@ -849,18 +991,10 @@ class TestQueryExpansion(unittest.TestCase):
             "lines[IDX_LINE].tanks[2@IDX_TANK].vol_outlet_flow",
         ]
         mpaths = [HubitModelPath(mpath) for mpath in mpaths]
-        _QueryExpansion(qpath, mpaths)
-
-        # Inconsistent query and mpaths
-        qpath = HubitQueryPath("lines[1].tanks[1].vol_outlet_flow")
-        mpaths = [
-            "lines[IDX_LINE].tanks[0@IDX_TANK].vol_outlet_flow",
-            "lines[IDX_LINE].tanks[1@IDX_TANK].vol_outlet_flow",
-            "lines[IDX_LINE].tanks[2@IDX_TANK].vol_outlet_flow",
-        ]
-        mpaths = [HubitModelPath(mpath) for mpath in mpaths]
-        with pytest.raises(HubitModelQueryError):
-            _QueryExpansion(qpath, mpaths)
+        # Use a dummy tree. It will not be used since the path normalization is bypassed
+        tree = DummyLengthTree()
+        cmps = get_mock_components(mpaths)
+        _QueryExpansion(qpath, mpaths, tree, cmps)
 
         # mpaths cannot have different index contexts
         qpath = HubitQueryPath("lines[:].tanks[:].vol_outlet_flow")
@@ -869,24 +1003,18 @@ class TestQueryExpansion(unittest.TestCase):
             "lines[IDX_LINE].tanks[1@IDX_OTHER].vol_outlet_flow",
         ]
         mpaths = [HubitModelPath(mpath) for mpath in mpaths]
+        cmps = get_mock_components(mpaths)
         with pytest.raises(HubitModelQueryError):
-            _QueryExpansion(qpath, mpaths)
+            _QueryExpansion(qpath, mpaths, tree, cmps)
 
         # No mpaths i.e. no provider
         qpath = HubitQueryPath("lines[:].tanks[:].vol_outlet_flow")
         mpaths = []
+        cmps = get_mock_components(mpaths)
         with pytest.raises(HubitModelQueryError):
-            _QueryExpansion(qpath, mpaths)
+            _QueryExpansion(qpath, mpaths, tree, cmps)
 
-    def _get_qexp_and_tree(make_tree_invalid: bool = False):
-        qpath = HubitQueryPath("lines[:].tanks[:].vol_outlet_flow")
-        mpaths = [
-            "lines[IDX_LINE].tanks[0@IDX_TANK].vol_outlet_flow",
-            "lines[IDX_LINE].tanks[1@IDX_TANK].vol_outlet_flow",
-            "lines[IDX_LINE].tanks[2@IDX_TANK].vol_outlet_flow",
-        ]
-        mpaths = [HubitModelPath(mpath) for mpath in mpaths]
-        qexp = _QueryExpansion(qpath, mpaths)
+    def _get_qexp():
 
         path = HubitModelPath("lines[IDX_LINE].tanks[0@IDX_TANK].vol_outlet_flow")
         yml_input = """
@@ -903,26 +1031,33 @@ class TestQueryExpansion(unittest.TestCase):
         """
         input_data = yaml.load(yml_input, Loader=yaml.FullLoader)
 
-        if make_tree_invalid:
-            # Delete tank so the tank list only has 2 elements. qexp expects at least 3
-            del input_data["lines"][0]["tanks"][-1]
+        tree = LengthTree.from_data(path, input_data)
 
-        return qexp, LengthTree.from_data(path, input_data)
+        qpath = HubitQueryPath("lines[:].tanks[:].vol_outlet_flow")
+        mpaths = [
+            "lines[IDX_LINE].tanks[0@IDX_TANK].vol_outlet_flow",
+            "lines[IDX_LINE].tanks[1@IDX_TANK].vol_outlet_flow",
+            "lines[IDX_LINE].tanks[2@IDX_TANK].vol_outlet_flow",
+        ]
+        mpaths = [HubitModelPath(mpath) for mpath in mpaths]
+        cmps = get_mock_components(mpaths)
+
+        qexp = _QueryExpansion(qpath, mpaths, tree, cmps)
+
+        return qexp
 
     def test_validate_tree(self):
-        qexp, tree = TestQueryExpansion._get_qexp_and_tree()
-        qexp.validate_tree(tree)
+        qexp = TestQueryExpansion._get_qexp()
+        qexp._validate_tree()
 
         # DummyTree passes validation
         path = HubitModelPath("segments.layers.positions")
-        tree = LengthTree.from_data(path, {})
-        qexp.validate_tree(tree)
+        qexp.tree = LengthTree.from_data(path, {})
+        qexp._validate_tree()
 
     def test_validate_tree_fail(self):
-        qexp, tree = TestQueryExpansion._get_qexp_and_tree(make_tree_invalid=True)
 
-        with pytest.raises(Exception):
-            qexp.validate_tree(tree)
+        qexp = TestQueryExpansion._get_qexp()
 
         # Tree corresponds to something 'segments[IDX_SEG].layers[IDX_LAY]'
         # while qexp was decomposed for the identifier 'IDX_TANK'
@@ -942,9 +1077,138 @@ class TestQueryExpansion(unittest.TestCase):
         """
         input_data = yaml.load(yml_input, Loader=yaml.FullLoader)
 
-        tree = LengthTree.from_data(path, input_data)
-        with pytest.raises(Exception):
-            qexp.validate_tree(tree)
+        qexp.tree = LengthTree.from_data(path, input_data)
+        with pytest.raises(HubitError):
+            qexp._validate_tree()
+
+    @staticmethod
+    def _get_tests():
+
+        # Test1
+        mpaths = [
+            "sites[IDX_SITE].lines[IDX_LINE].tanks[IDX_TANK].Q_yield",
+        ]
+        cmps = get_mock_components(mpaths)
+        tests = [(mpaths, cmps)]
+        # Test 2
+        mpaths = [
+            "sites[IDX_SITE].lines[IDX_LINE].tanks[0@IDX_TANK].Q_yield",
+            "sites[IDX_SITE].lines[IDX_LINE].tanks[1@IDX_TANK].Q_yield",
+            "sites[IDX_SITE].lines[IDX_LINE].tanks[2:@IDX_TANK].Q_yield",
+        ]
+        # Explicitly defining tank three doesn't work since it doesn't exits for the first site
+        # "sites[IDX_SITE].lines[IDX_LINE].tanks[2@IDX_TANK].Q_yield",
+        # "sites[IDX_SITE].lines[IDX_LINE].tanks[3@IDX_TANK].Q_yield",
+        cmps = get_mock_components(mpaths)
+
+        tests.append((mpaths, cmps))
+        return tests
+
+    def test_collect_results(self):
+        """
+        Test that the flat results are collected correctly when forming the response
+
+        - 1 provider -> 1 mpath
+        - 1 production site
+        """
+        tree, flat_results = _get_data()
+        test_items = TestQueryExpansion._get_tests()
+
+        for mpaths, cmps in test_items:
+            mpaths = [HubitModelPath(mpath) for mpath in mpaths]
+            with self.subTest(mpaths=mpaths, cmps=cmps):
+                # Positive indices
+                qpath = HubitQueryPath("sites[0].lines[0].tanks[0].Q_yield")
+                _tree = tree.prune_from_path(qpath, inplace=False)
+                qexp = _QueryExpansion(qpath, mpaths, _tree, cmps)
+                result = qexp.collect_results(flat_results)
+                # Query has two index wildcards hence nested list
+                assert result == 10.0
+
+                # Positive indices
+                qpath = HubitQueryPath("sites[1].lines[0].tanks[3].Q_yield")
+                _tree = tree.prune_from_path(qpath, inplace=False)
+                qexp = _QueryExpansion(qpath, mpaths, _tree, cmps)
+                result = qexp.collect_results(flat_results)
+                # Query has two index wildcards hence nested list
+                assert result == 1.0
+
+                # Negative indices
+                qpath = HubitQueryPath("sites[-1].lines[-1].tanks[-1].Q_yield")
+                qpath = HubitQueryPath("sites[1].lines[0].tanks[-1].Q_yield")
+                _tree = tree.prune_from_path(qpath, inplace=False)
+                qexp = _QueryExpansion(qpath, mpaths, _tree, cmps)
+                result = qexp.collect_results(flat_results)
+                # Query has two index wildcards hence nested list
+                assert result == 1.0
+
+                # Negative index
+                qpath = HubitQueryPath("sites[:].lines[:].tanks[-1].Q_yield")
+                _tree = tree.prune_from_path(qpath, inplace=False)
+                qexp = _QueryExpansion(qpath, mpaths, _tree, cmps)
+                result = qexp.collect_results(flat_results)
+                # Query has two index wildcards hence nested list
+                assert result == [[4.0], [1.0]]
+
+                # Positive index
+                qpath = HubitQueryPath("sites[:].lines[:].tanks[2].Q_yield")
+                _tree = tree.prune_from_path(qpath, inplace=False)
+                qexp = _QueryExpansion(qpath, mpaths, _tree, cmps)
+                result = qexp.collect_results(flat_results)
+                # Query has two index wildcards hence nested list
+                assert result == [[4.0], [3.0]]
+
+                # All tanks on all lines
+                qpath = HubitQueryPath("sites[0].lines[:].tanks[:].Q_yield")
+                _tree = tree.prune_from_path(qpath, inplace=False)
+                qexp = _QueryExpansion(qpath, mpaths, _tree, cmps)
+                result = qexp.collect_results(flat_results)
+                # Query has two index wildcards hence nested list
+                assert result == [[10.0, 6.0, 4.0]]
+
+                # All tanks on all lines in all sites
+                qpath = HubitQueryPath("sites[:].lines[:].tanks[:].Q_yield")
+                _tree = tree.prune_from_path(qpath, inplace=False)
+                qexp = _QueryExpansion(qpath, mpaths, _tree, cmps)
+                result = qexp.collect_results(flat_results)
+                # Query has two index wildcards hence nested list
+                assert result == [[[10.0, 6.0, 4.0]], [[20.0, 8.0, 3.0, 1.0]]]
+
+    def test_filter_mpaths_for_qpath_index_ranges(self):
+        qpath = HubitQueryPath("sites[0].lines[0].tanks[:].Q_yield")
+        mpaths = [
+            HubitModelPath(
+                "sites[0@IDX_SITE].lines[0@IDX_LINE].tanks[IDX_TANK].Q_yield"
+            ),
+            HubitModelPath(
+                "sites[0@IDX_SITE].lines[0@IDX_LINE].tanks[IDX_TANK].Q_yield"
+            ),
+            HubitModelPath(
+                "sites[0@IDX_SITE].lines[1@IDX_LINE].tanks[IDX_TANK].Q_yield"
+            ),
+        ]
+
+        # Index scpoe for each mpath
+        index_scopes = [
+            {"IDX_TANK": PathIndexRange("0")},
+            {"IDX_TANK": PathIndexRange("1:")},
+            {"IDX_TANK": PathIndexRange(":")},
+        ]
+        tree, _ = _get_data()
+        pruned_tree = tree.prune_from_path(qpath, inplace=False)
+
+        result = _QueryExpansion._filter_mpaths_for_qpath_index_ranges(
+            qpath,
+            mpaths,
+            index_scopes,
+            pruned_tree,
+        )
+
+        expected_result = [
+            mpath.set_range_for_idxid(index_scope)
+            for mpath, index_scope in zip(mpaths[:2], index_scopes[:2])
+        ]
+        assert result == expected_result
 
 
 if __name__ == "__main__":
