@@ -24,7 +24,6 @@ class TestWorker(unittest.TestCase):
         """
         Fails since query does not match.
         """
-        qrunner = None
         func = None
         version = None
         cfg = {
@@ -44,7 +43,8 @@ class TestWorker(unittest.TestCase):
         querystring = HubitQueryPath("shared.attr.path")
         with self.assertRaises(HubitWorkerError) as context:
             w = _Worker(
-                qrunner,
+                lambda x: x,
+                lambda x: x,
                 component,
                 querystring,
                 func,
@@ -57,7 +57,6 @@ class TestWorker(unittest.TestCase):
         """
         Initialize a simple worker with no idxids
         """
-        qrunner = None
         func = None
         version = None
         cfg = {
@@ -77,7 +76,8 @@ class TestWorker(unittest.TestCase):
         # Query something known to exist
         querystring = HubitQueryPath(component.provides_results[0].path)
         w = _Worker(
-            qrunner,
+            lambda x: x,
+            lambda x: x,
             component,
             querystring,
             func,
@@ -88,8 +88,6 @@ class TestWorker(unittest.TestCase):
 
     def _make_worker():
         qrunner = Mock()
-        qrunner._set_worker
-        qrunner._set_worker_working
         qrunner.check_cache.return_value = None
         cname = None
         func = dummy_function
@@ -130,7 +128,8 @@ class TestWorker(unittest.TestCase):
         _tree_for_idxcontext = tree_for_idxcontext([component], inputdata)
 
         w = _Worker(
-            qrunner,
+            lambda x: x,
+            lambda x: x,
             component,
             querystring,
             func,
@@ -141,12 +140,6 @@ class TestWorker(unittest.TestCase):
 
         return w
 
-    def test_set_results_id(self):
-        w = TestWorker._make_worker()
-        result = w.set_results_id(["A", "B", "C"])
-        expected_result = "9fe92bbe331cb58741ebc0256f3eb32c"
-        assert result == expected_result
-
     def test_4(self):
         """
         Adding required data to worker stepwise to see that it
@@ -228,10 +221,10 @@ class TestWorker(unittest.TestCase):
                 w.is_ready_to_work() == (len(pending_results_paths) == 0)
             )
 
-        print("WWW", w.pending_results_paths, pending_results_paths)
-        print("WWW", w.pending_input_paths, pending_input_paths)
+        # Start work
+        w.work()
 
-        # After adding last attribute the worker starts running (sequentially)
+        # Work is sequential so now the results are ready
         self.assertTrue(w.results_ready())
 
         self.assertTrue(all(itests_paths_pending))
@@ -247,7 +240,6 @@ class TestWorker(unittest.TestCase):
         Initialize worker with ILOC locations in
         query and ILOC wildcards in bindings
         """
-        qrunner = None
         func = None
         version = None
         cfg = """
@@ -274,11 +266,12 @@ class TestWorker(unittest.TestCase):
         nodes.extend(lay_nodes)
         level_names = "IDX_SEG", "IDX_LAY"
         tree = LengthTree(nodes, level_names)
-        _tree_for_idxcontext = {tree.get_idx_context(): tree}
+        _tree_for_idxcontext = {tree.index_context: tree}
 
         querystr = HubitQueryPath(querystr)
         w = _Worker(
-            qrunner,
+            lambda x: x,
+            lambda x: x,
             component,
             querystr,
             func,
@@ -364,7 +357,6 @@ class TestWorker(unittest.TestCase):
         """
         Test compression of indices. The query does not include any indices
         """
-        qrunner = None
         func = None
         version = None
         comp_yml = """
@@ -389,11 +381,12 @@ class TestWorker(unittest.TestCase):
         tree = LengthTree(nodes, level_names)
         dummy_tree = DummyLengthTree()
 
-        _tree_for_idxcontext = {"": dummy_tree, tree.get_idx_context(): tree}
+        _tree_for_idxcontext = {"": dummy_tree, tree.index_context: tree}
 
         querystr = HubitQueryPath(querystr)
         w = _Worker(
-            qrunner,
+            lambda x: x,
+            lambda x: x,
             component,
             querystr,
             func,
